# -*- coding: utf-8 -*-

import os
from datetime import datetime,timedelta
from configparser import ConfigParser
import dask
import dask.dataframe
import dask.multiprocessing
from dask.diagnostics import ProgressBar
import numpy as np
from numpy.lib.function_base import iterable
from processor import DldProcessor
from processor.utilities import misc
from processor.pah import BeamtimeDaqAccess
import json

_VERBOSE = False

# Try to load the Cython version of the microbunch assignment code
# If fails, load a vanilla python equivalent (Steinn Y. Agustsson)
try:
    import processor.cscripts.DldFlashProcessorCy as DldFlashProcessorCy

    if _VERBOSE:
        print('loaded cython module')
except ImportError as e:
    print('Failed loading Cython script. Using Python version instead. TODO: FIX IT!!#n Error msg: {}'.format(e))
    import processor.cscripts.DldFlashProcessorNotCy as DldFlashProcessorCy

assignToMircobunch = DldFlashProcessorCy.assignToMircobunch


class DldFlashProcessor(DldProcessor.DldProcessor):
    """  
    The class reads an existing run and allows to generated binned multidimensional arrays,
    which can be used directly or saved as HDF5 dataframes.
    
    This class reads an existing run and generates a hdf5 file containing the dask dataframes.
    It is intended to be used with data generated from August 31, 2017 to September 19, 2017.
    Version 4 enables read out of macrobunch ID. For evaluation the start ID is set to zero.
    
    **Version 5**
    
    * introduces overwriting of PAH classes for correct handling of macrobunchID
    * introduces writeRunToMultipleParquet function, for parquet file generation on machines with low ram
    * changed variables to class variables for easier implementation on different machines
    * added some print functions with information about the run that is being imported.

    Had to change the delay stage channel, as the old one (.../ENC.DELAY) stored groups of ~ 10 times the same value
    (is probably read out with 10 Hz. The new channel is the column (index!) one of .../ENC.
    This change makes the treatment of problematic runs obsolete.


    **Attributes**\n
    runNumber: int
        number of the run from which data is taken.
    pulseIdInterval: (int, int) tuple
        macrobunch ID corresponding to the interval of data read from the given run.
    dd: dask dataframe
        dataframe containing chosen channel information from the given run
    dd_microbunch: dask dataframe
        dataframe containing chosen channel information from the given run.
    """

    def __init__(self, runNumber=None, pulseIdInterval=None, settings=None):

        super().__init__(settings)

        self.runNumber = runNumber
        self.pulseIdInterval = pulseIdInterval

    def readData(self, runNumber=None, pulseIdInterval=None, path=None):
        """Read data by run number or macrobunch pulseID interval.

        Useful for scans that would otherwise hit the machine's memory limit.

        **Parameters**\n
        runNumber: int | None (default to ``self.runNumber``)
            number of the run from which to read data. If None, requires pulseIdInterval.
        pulseIdInterval: (int, int) | None (default to ``self.pulseIdInterval``)
            first and last macrobunches of selected data range. If None, the whole run
            defined by runNumber will be taken.
        path: str | None (default to ``self.DATA_RAW_DIR``)
            path to location where raw HDF5 files are stored.

        This is a union of the readRun and readInterval methods defined in previous versions.
        """

        # Update instance attributes based on input parameters
        if runNumber is None:
            runNumber = self.runNumber
        else:
            self.runNumber = runNumber

        if pulseIdInterval is None:
            pulseIdInterval = self.pulseIdInterval
        else:
            self.pulseIdInterval = pulseIdInterval

        if (pulseIdInterval is None) and (runNumber is None):
            raise ValueError('Need either runNumber or pulseIdInterval to know what data to read.')

        if path is not None:
            try:
                daqAccess = BeamtimeDaqAccess.create(path)
            except:
                self.path_to_run = misc.get_path_to_run(runNumber, path)
                daqAccess = BeamtimeDaqAccess.create(self.path_to_run)
        else:
            path = self.DATA_RAW_DIR
            self.path_to_run = misc.get_path_to_run(runNumber, path)
            daqAccess = BeamtimeDaqAccess.create(self.path_to_run)

        self.daqAddresses = []
        self.pulseIdInterval = self.getIds(runNumber, path)
        # Parse the settings file in the DAQ channels section for the list of
        # h5 addresses to read from raw and add to the dataframe.
        print('loading data...')

        for name, entry in self.settings['DAQ channels'].items():
            name = misc.camelCaseIt(name)
            val = str(entry)
            if daqAccess.isChannelAvailable(val, self.pulseIdInterval):
                self.daqAddresses.append(name)
                if _VERBOSE:
                    print('assigning address: {}: {}'.format(name.ljust(20), val))
                setattr(self, name, val)
            else:
                # if _VERBOSE:
                print('skipping address missing from data: {}: {}'.format(name.ljust(20), val))

        # TODO: get the available pulse id from PAH
        if pulseIdInterval is None:
            print('Reading DAQ data from run {}... Please wait...'.format(runNumber))

            for address_name in self.daqAddresses:
                if _VERBOSE:
                    print('reading address: {}'.format(address_name))
                try:
                    attrVal = getattr(self, address_name)
                    values, otherStuff = daqAccess.allValuesOfRun(attrVal, runNumber)
                except AssertionError:
                    print('Assertion error: {}'.format(address_name, attrVal, values, otherStuff))

                setattr(self, address_name, values)
                if address_name == 'macroBunchPulseId':  # catch the value of the first macrobunchID
                    pulseIdInterval = (otherStuff[0], otherStuff[-1])
                    self.pulseIdInterval = pulseIdInterval
                    macroBunchPulseId_correction = pulseIdInterval[0]

                if address_name == 'timeStamp':  # catch the time stamps
                    startEndTime = (values[0, 0], values[-1, 0])
                    self.startEndTime = startEndTime

            numOfMacrobunches = pulseIdInterval[1] - pulseIdInterval[0]



        else:
            print('reading DAQ data from interval {}'.format(pulseIdInterval))
            self.pulseIdInterval = pulseIdInterval
            for address_name in self.daqAddresses:
                if _VERBOSE:
                    print('reading address: {}'.format(address_name))
                values = daqAccess.valuesOfInterval(getattr(self, address_name), pulseIdInterval)
                setattr(self, address_name, values)
                if address_name == 'timeStamp':  # catch the time stamps
                    startEndTime = (values[0, 0], values[-1, 0])
                    self.startEndTime = startEndTime
            numOfMacrobunches = pulseIdInterval[1] - pulseIdInterval[0]
            macroBunchPulseId_correction = pulseIdInterval[0]

        # necessary corrections for specific channels:
        try:
            self.delayStage = self.delayStage[:, 1]
        except:
            try:
                self.delayStage = self.delayStage[:, 0]
                print('1030nm Laser')
            except:
                print('no delay stage')
        if self.CORRECT_MB_ID:
            self.macroBunchPulseId -= macroBunchPulseId_correction
        self.dldMicrobunchId -= self.UBID_OFFSET

        if _VERBOSE:
            print('Counting electrons...')

        electronsToCount = self.dldPosX.copy().flatten()
        electronsToCount = np.nan_to_num(electronsToCount)
        electronsToCount = electronsToCount[electronsToCount > 0]
        electronsToCount = electronsToCount[electronsToCount < 10000]
        self.numOfElectrons = len(electronsToCount)
        self.electronsPerMacrobunch = int(self.numOfElectrons / numOfMacrobunches)

        self.runInfo = {
            'runNumber': self.runNumber,
            'pulseIdInterval': self.pulseIdInterval,
            'numberOfMacrobunches': numOfMacrobunches,
            'numberOfElectrons': self.numOfElectrons,
            'electronsPerMacrobunch': self.electronsPerMacrobunch,
        }
        try:
            self.runInfo['timestampStart'] = int(self.startEndTime[0])
            self.runInfo['timestampStop'] = int(self.startEndTime[1])
            self.runInfo['timestampDuration'] = int(self.startEndTime[1] - self.startEndTime[0])
            self.runInfo['timeStart'] = datetime.utcfromtimestamp(self.startEndTime[0]).strftime('%Y-%m-%d %H:%M:%S')
            self.runInfo['timeStop'] = datetime.utcfromtimestamp(self.startEndTime[1]).strftime('%Y-%m-%d %H:%M:%S')
            self.runInfo['timeDuration'] = str(timedelta(seconds=int(self.startEndTime[1] - self.startEndTime[0])))
        except:
            self.runInfo['timestampStart'] = None
            self.runInfo['timestampStop'] = None
            self.runInfo['timestampDuration'] = None
            self.runInfo['timeStart'] = None
            self.runInfo['timeStop'] = None
            self.runInfo['timeDuration'] = None

        self.printRunOverview()

        # Old Print style
        # print('Run {0} contains {1:,} Macrobunches, from {2:,} to {3:,}' \
        #       .format(runNumber, numOfMacrobunches, pulseIdInterval[0], pulseIdInterval[1]))
        # try:
        #     print("start time: {}, end time: {}, total time: {}"
        #           .format(datetime.utcfromtimestamp(startEndTime[0]).strftime('%Y-%m-%d %H:%M:%S'),
        #                   datetime.utcfromtimestamp(startEndTime[1]).strftime('%Y-%m-%d %H:%M:%S'),
        #                   datetime.utcfromtimestamp(startEndTime[1] - startEndTime[0]).strftime('%H:%M:%S')))
        # except:
        #     pass
        #
        # print("Number of electrons: {0:,}; {1:,} e/Mb ".format(self.numOfElectrons, self.electronsPerMacrobunch))
        if not bool(self.metadata):
            self.metadata = self.update_metadata()

        print("Creating dataframes... Please wait...")
        with ProgressBar():
            self.createDataframePerElectron()
            print('Electron dataframe created.')
            self.createDataframePerMicrobunch()
            print('Microbunch dataframe created.')
            print('Reading Complete.')

    def readData_old(self, runNumber=None, pulseIdInterval=None, path=None):
        """Read data by run number or macrobunch pulseID interval.

        Useful for scans that would otherwise hit the machine's memory limit.

        **Parameters**\n
        runNumber: int | None (default to ``self.runNumber``)
            number of the run from which to read data. If None, requires pulseIdInterval.
        pulseIdInterval: (int, int) | None (default to ``self.pulseIdInterval``)
            first and last macrobunches of selected data range. If None, the whole run
            defined by runNumber will be taken.
        path: str | None (default to ``self.DATA_RAW_DIR``)
            path to location where raw HDF5 files are stored.

        This is a union of the readRun and readInterval methods defined in previous versions.
        """

        # Update instance attributes based on input parameters
        if runNumber is None:
            runNumber = self.runNumber
        else:
            self.runNumber = runNumber

        if pulseIdInterval is None:
            pulseIdInterval = self.pulseIdInterval
        else:
            self.pulseIdInterval = pulseIdInterval

        if (pulseIdInterval is None) and (runNumber is None):
            raise ValueError('Need either runNumber or pulseIdInterval to know what data to read.')

        print('searching for data...')
        if path is not None:
            try:
                daqAccess = BeamtimeDaqAccess.create(path)
            except:
                self.path_to_run = misc.get_path_to_run(runNumber, path)
                daqAccess = BeamtimeDaqAccess.create(self.path_to_run)
        else:
            path = self.DATA_RAW_DIR
            self.path_to_run = misc.get_path_to_run(runNumber, path)
            daqAccess = BeamtimeDaqAccess.create(self.path_to_run)

        self.daqAddresses = []
        # Parse the settings file in the DAQ channels section for the list of
        # h5 addresses to read from raw and add to the dataframe.
        for name, entry in self.settings['DAQ channels'].items():
            name = misc.camelCaseIt(name)
            val = str(entry)
            if daqAccess.isChannelAvailable(val, self.getIds(runNumber, path)):
                self.daqAddresses.append(name)
                if _VERBOSE:
                    print('assigning address: {}: {}'.format(name.ljust(20), val))
                setattr(self, name, val)
            else:
                # if _VERBOSE:
                print('skipping address missing from data: {}: {}'.format(name.ljust(20), val))

        # TODO: get the available pulse id from PAH
        if pulseIdInterval is None:
            print('Reading DAQ data from run {}... Please wait...'.format(runNumber))

            for address_name in self.daqAddresses:
                if _VERBOSE:
                    print('reading address: {}'.format(address_name))
                try:
                    attrVal = getattr(self, address_name)
                    values, otherStuff = daqAccess.allValuesOfRun(attrVal, runNumber)
                except AssertionError:
                    print('Assertion error: {}'.format(address_name, attrVal, values, otherStuff))

                setattr(self, address_name, values)
                if address_name == 'macroBunchPulseId':  # catch the value of the first macrobunchID
                    pulseIdInterval = (otherStuff[0], otherStuff[-1])
                    self.pulseIdInterval = pulseIdInterval
                    macroBunchPulseId_correction = pulseIdInterval[0]

                if address_name == 'timeStamp':  # catch the time stamps
                    startEndTime = (values[0, 0], values[-1, 0])
                    self.startEndTime = startEndTime

            numOfMacrobunches = pulseIdInterval[1] - pulseIdInterval[0]
            print('Run {0} contains {1:,} Macrobunches, from {2:,} to {3:,}' \
                  .format(runNumber, numOfMacrobunches, pulseIdInterval[0], pulseIdInterval[1]))
            try:
                print("start time: {}, end time: {}, total time: {}"
                      .format(datetime.utcfromtimestamp(startEndTime[0]).strftime('%Y-%m-%d %H:%M:%S'),
                              datetime.utcfromtimestamp(startEndTime[1]).strftime('%Y-%m-%d %H:%M:%S'),
                              datetime.utcfromtimestamp(startEndTime[1] - startEndTime[0]).strftime('%H:%M:%S')))
            except:
                pass
        else:
            print('reading DAQ data from interval {}'.format(pulseIdInterval))
            self.pulseIdInterval = pulseIdInterval
            for address_name in self.daqAddresses:
                if _VERBOSE:
                    print('reading address: {}'.format(address_name))
                setattr(self, address_name, daqAccess.valuesOfInterval(getattr(self, address_name), pulseIdInterval))
            numOfMacrobunches = pulseIdInterval[1] - pulseIdInterval[0]
            macroBunchPulseId_correction = pulseIdInterval[0]

        # necessary corrections for specific channels:
        self.delayStage = self.delayStage[:, 1]
        self.macroBunchPulseId -= macroBunchPulseId_correction
        self.dldMicrobunchId -= self.UBID_OFFSET

        if _VERBOSE:
            print('Counting electrons...')

        electronsToCount = self.dldPosX.copy().flatten()
        electronsToCount = np.nan_to_num(electronsToCount)
        electronsToCount = electronsToCount[electronsToCount > 0]
        electronsToCount = electronsToCount[electronsToCount < 10000]
        self.numOfElectrons = len(electronsToCount)
        self.electronsPerMacrobunch = int(self.numOfElectrons / numOfMacrobunches)
        print("Number of electrons: {0:,}; {1:,} e/Mb ".format(self.numOfElectrons, self.electronsPerMacrobunch))

        print("Creating dataframes... Please wait...")
        pbar = ProgressBar()
        with pbar:
            self.createDataframePerElectron()
            print('Electron dataframe created.')
            self.createDataframePerMicrobunch()
            print('Microbunch dataframe created.')

    def createDataframePerElectronRange(self, mbIndexStart, mbIndexEnd):
        """ Create a numpy array indexed by photoelectron events for a given range,
        [start, end), of electron macrobunch IDs.
        
        **Parameters**\n
        mbIndexStart: int
            The starting (inclusive) macrobunch ID.
        mbIndexEnd: int
            The ending (non-inclusive) macrobunch ID.
        
        **Return**\n
        da: numpy array
            Indexed photoelectron events.
        """
        # the chunk size here is too large in order to do the chunking by the loop around it.

        # Here all the columns to be stored in the dd dataframe are created from the raw h5 file
        # Each columns requires an ad hoc treatment, so they all need to be done individually
        arrayCols = []  # TODO: find less ad hoc solution
        colNames = []

        if 'dldPosX' in self.daqAddresses:
            daX = self.dldPosX[mbIndexStart:mbIndexEnd, :].flatten()
            arrayCols.append(daX)
            colNames.append('dldPosX')

        if 'dldPosY' in self.daqAddresses:
            daY = self.dldPosY[mbIndexStart:mbIndexEnd, :].flatten()
            arrayCols.append(daY)
            colNames.append('dldPosY')

        if 'dldTime' in self.daqAddresses:
            daTime = self.dldTime[mbIndexStart:mbIndexEnd, :].flatten()
            arrayCols.append(daTime)
            colNames.append('dldTime')

        if 'dldAux0' in self.daqAddresses:
            dldAuxChannels = {'sampleBias': 0,
                              'tofVoltage': 1,
                              'extractorVoltage': 2,
                              'extractorCurrent': 3,
                              'cryoTemperature': 4,
                              'sampleTemperature': 5,
                              'dldTimeBinSize': 15,
                              }

            for name, chan in dldAuxChannels.items():
                da = np.zeros_like(self.dldMicrobunchId[mbIndexStart:mbIndexEnd, :])
                da[:, :] = (self.dldAux0[mbIndexStart:mbIndexEnd, chan])[:, None]
                da = da.flatten()
                arrayCols.append(da)
                colNames.append(name)
            # daSampleBias = np.zeros_like(self.dldMicrobunchId[mbIndexStart:mbIndexEnd, :])
            # daSampleBias[:, :] = (self.dldAux0[mbIndexStart:mbIndexEnd, 0])[:, None]
            # daSampleBias = daSampleBias.flatten()
            # arrayCols.append(daSampleBias)
            # colNames.append('sampleBias')
            #
            # daTofVoltage = np.zeros_like(self.dldMicrobunchId[mbIndexStart:mbIndexEnd, :])
            # daTofVoltage[:, :] = (self.dldAux0[mbIndexStart:mbIndexEnd, 1])[:, None]
            # daTofVoltage = daTofVoltage.flatten()
            # arrayCols.append(daTofVoltage)
            # colNames.append('tofVoltage')
            #
            # daExtractorVoltage = np.zeros_like(self.dldMicrobunchId[mbIndexStart:mbIndexEnd, :])
            # daExtractorVoltage[:, :] = (self.dldAux0[mbIndexStart:mbIndexEnd, 2])[:, None]
            # daExtractorVoltage = daExtractorVoltage.flatten()
            # arrayCols.append(daExtractorVoltage)
            # colNames.append('extractorVoltage')

        if 'delayStage' in self.daqAddresses:
            delayStageArray = np.zeros_like(self.dldMicrobunchId[mbIndexStart:mbIndexEnd, :])
            delayStageArray[:, :] = (self.delayStage[mbIndexStart:mbIndexEnd])[:, None]
            daDelaystage = delayStageArray.flatten()
            arrayCols.append(daDelaystage)
            colNames.append('delayStage')

        if 'streakCam' in self.daqAddresses:
            streakCamArray = np.zeros_like(self.dldMicrobunchId[mbIndexStart:mbIndexEnd, :])
            streakCamArray[:, :] = (self.streakCam[mbIndexStart:mbIndexEnd, 0])[:, None]
            daStreakCam = streakCamArray.flatten()
            arrayCols.append(daStreakCam)
            colNames.append('streakCamera')

        if 'bam' in self.daqAddresses:
            bamArray = assignToMircobunch(
                self.dldMicrobunchId[mbIndexStart:mbIndexEnd, :].astype(np.float64),
                self.bam[mbIndexStart:mbIndexEnd, :].astype(np.float64))
            daBam = bamArray.flatten()
            arrayCols.append(daBam)
            colNames.append('bam')

        if 'dldMicrobunchId' in self.daqAddresses:
            daMicrobunchId = self.dldMicrobunchId[mbIndexStart:mbIndexEnd, :].flatten()
            arrayCols.append(daMicrobunchId)
            colNames.append('dldMicrobunchId')

        if 'dldDetectorId' in self.daqAddresses:
            dldDetectorId = (self.dldDetectorId[mbIndexStart:mbIndexEnd, :].copy()).astype(int) % 2
            daDetectorId = dldDetectorId.flatten()
            arrayCols.append(daDetectorId)
            colNames.append('dldDetectorId')

        if 'dldSectorId' in self.daqAddresses:
            dldSectorId = (self.dldSectorId[mbIndexStart:mbIndexEnd, :].copy()).astype(int) % np.power(2,self.DLD_ID_BITS)
            daSectorId = dldSectorId.flatten()
            arrayCols.append(daSectorId)
            colNames.append('dldSectorId')

        if 'bunchCharge' in self.daqAddresses:
            bunchChargeArray = assignToMircobunch(
                self.dldMicrobunchId[mbIndexStart:mbIndexEnd, :].astype(np.float64),
                self.bunchCharge[mbIndexStart:mbIndexEnd, :].astype(np.float64))
            daBunchCharge = bunchChargeArray.flatten()
            arrayCols.append(daBunchCharge)
            colNames.append('bunchCharge')

        if 'opticalDiode' in self.daqAddresses:
            opticalDiodeArray = assignToMircobunch(
                self.dldMicrobunchId[mbIndexStart:mbIndexEnd, :].astype(np.float64),
                self.opticalDiode[mbIndexStart:mbIndexEnd, :].astype(np.float64))
            daOpticalDiode = opticalDiodeArray.flatten()
            arrayCols.append(daOpticalDiode)
            colNames.append('opticalDiode')

        if 'gmdTunnel' in self.daqAddresses:
            gmdTunnelArray = assignToMircobunch(
                self.dldMicrobunchId[mbIndexStart:mbIndexEnd, :].astype(np.float64),
                self.gmdTunnel[mbIndexStart:mbIndexEnd, :].astype(np.float64))
            daGmdTunnel = gmdTunnelArray.flatten()
            arrayCols.append(daGmdTunnel)
            colNames.append('gmdTunnel')

        if 'gmdBda' in self.daqAddresses:
            gmdBdaArray = assignToMircobunch(
                self.dldMicrobunchId[mbIndexStart:mbIndexEnd, :].astype(np.float64),
                self.gmdBda[mbIndexStart:mbIndexEnd, :].astype(np.float64))
            daGmdBda = gmdBdaArray.flatten()
            arrayCols.append(daGmdBda)
            colNames.append('gmdBda')

        if 'monochromatorActEnergy' in self.daqAddresses:
            monochromatorChannels = {'delta1': 0,
                              'delta2': 1,
                              'mirrorAngle': 2,
                              'gratingAngle': 3,
                              }

            alpha = 2 * self.monochromatorActEnergy[mbIndexStart:mbIndexEnd,monochromatorChannels['delta1']] + 90 -\
                    self.monochromatorActEnergy[mbIndexStart:mbIndexEnd,monochromatorChannels['delta2']]
            beta = -1 * self.monochromatorActEnergy[mbIndexStart:mbIndexEnd,monochromatorChannels['delta2']] - 86
            GratingDensity = 200
            DiffrOrder = 1

            da = np.zeros_like(self.dldMicrobunchId[mbIndexStart:mbIndexEnd, :])
            da[:, :] = np.array(1239.84/ ((np.sin(beta / 180 * np.pi)+np.sin(alpha / 180 * np.pi)) * 1e9 / (DiffrOrder * GratingDensity * 1000)))[:,None]
            # for name, chan in monochromatorChannels.items():
            da = da.flatten()

            arrayCols.append(da)
            colNames.append('monochromatorActEnergy')



        if 'monochromatorPhotonEnergy' in self.daqAddresses:
            monochromatorPhotonEnergyArray = assignToMircobunch(
                self.dldMicrobunchId[mbIndexStart:mbIndexEnd, :].astype(np.float64),
                self.monochromatorPhotonEnergy[mbIndexStart:mbIndexEnd, :].astype(np.float64))
            daMonochromatorPhotonEnergy = monochromatorPhotonEnergyArray.flatten()
            # daMonochromatorPhotonEnergy = np.zeros_like(self.dldMicrobunchId[mbIndexStart:mbIndexEnd, :])
            # daMonochromatorPhotonEnergy[:, :] = (self.monochromatorPhotonEnergy[mbIndexStart:mbIndexEnd])[:, None]
            # daMonochromatorPhotonEnergy = daMonochromatorPhotonEnergy.flatten()
            # handle nans # TODO: This channel should become a static value
            nan = np.float64(np.nanmean(self.monochromatorPhotonEnergy))
            daMonochromatorPhotonEnergy = np.nan_to_num(daMonochromatorPhotonEnergy,nan=nan)
            arrayCols.append(daMonochromatorPhotonEnergy)
            colNames.append('monochromatorPhotonEnergy')

        if 'i0Monitor' in self.daqAddresses:
            def I0BunchTrain2Pulses():
                """
                Cut, reshape and average the ADC signal of the I0 monitor. 
                Preliminary procedure: migth be optimized and tunend in to the DAQ 
                """
                i0Cutted = np.array(
                    np.split(-self.i0Monitor[:, self.I0ID_OFFSET:self.I0ID_OFFSET + 108 * self.I0ID_N], self.I0ID_N,
                             axis=1))
                i0BG = i0Cutted[:, :, self.I0_MEAN_LOW:self.I0_MEAN_HIGH].mean(2)
                return np.nansum(i0Cutted[:, :, self.I0_SUM_LOW:self.I0_SUM_HIGH] - i0BG[:, :, None], 2).T

            i0Data = I0BunchTrain2Pulses()

            i0Array = assignToMircobunch(
                self.dldMicrobunchId[mbIndexStart:mbIndexEnd, :].astype(np.float64),
                i0Data[mbIndexStart:mbIndexEnd, :].astype(np.float64))
            daI0 = i0Array.flatten()
            arrayCols.append(daI0)
            colNames.append('i0Monitor')

        # convert the laser polarization motor position to the electron format
        if 'pumpPol' in self.daqAddresses:
            pumpPolArray = np.zeros_like(self.dldMicrobunchId[mbIndexStart:mbIndexEnd, :])
            pumpPolArray[:, :] = (self.pumpPol[mbIndexStart:mbIndexEnd, 0])[:, None]
            daPumpPol = pumpPolArray.flatten()
            arrayCols.append(daPumpPol)
            colNames.append('pumpPol')

        # convert the MacroBunchPulseId to the electron format. No check because this surely exists
        if 'macroBunchPulseId' in self.daqAddresses:
            macroBunchPulseIdArray = np.zeros_like(self.dldMicrobunchId[mbIndexStart:mbIndexEnd, :])
            macroBunchPulseIdArray[:, :] = (self.macroBunchPulseId[mbIndexStart:mbIndexEnd, 0])[:, None]
            daMacroBunchPulseId = macroBunchPulseIdArray.flatten()
            arrayCols.append(daMacroBunchPulseId)
            colNames.append('macroBunchPulseId')

        # convert the timeStamp to the electron format. No check because this surely exists
        if 'timeStamp' in self.daqAddresses:
            timeStampArray = np.zeros_like(self.dldMicrobunchId[mbIndexStart:mbIndexEnd, :])
            timeStampArray[:, :] = (self.timeStamp[mbIndexStart:mbIndexEnd, 0])[:, None]
            daTimeStamp = timeStampArray.flatten()
            arrayCols.append(daTimeStamp)
            colNames.append('timeStamp')

        # the Aux channel: aux0:
        # aux0Arr= assignToMircobunch(self.dldMicrobunchId[mbIndexStart:mbIndexEnd, :].astype(np.float64), self.dldAux[mbIndexStart:mbIndexEnd, 0].astype(np.float64))
        # daAux0 = dask.array.from_array(aux0Arr.flatten(), chunks=(chunks))

        # the Aux channel: aux1:
        # aux1Arr= assignToMircobunch(self.dldMicrobunchId[mbIndexStart:mbIndexEnd, :].astype(np.float64), self.dldAux[mbIndexStart:mbIndexEnd, 1].astype(np.float64))
        # daAux1 = dask.array.from_array(aux0Arr.flatten(), chunks=(chunks))

        # added macroBunchPulseId at last position
        # da = dask.array.stack([daX, daY, daTime, daDelaystage, daBam, daMicrobunchId,
        #                       daDetectorId, daSectorId, daBunchCharge, daOpticalDiode,
        #                       daGmdTunnel, daMacroBunchPulseId])
        da = np.stack(arrayCols)
        if len(self.ddColNames) != len(colNames):
            self.ddColNames = colNames
        return da

    def createDataframePerElectron(self):
        """ Create a dataframe indexed by photoelectron events from the read arrays
        (either from the test file or the run number). The method needs no input parameters.
        """

        # self.dldTime=self.dldTime*self.dldTimeStep
        if _VERBOSE:
            print('creating electron dataframe...')

        maxIndex = self.dldTime.shape[0]

        if self.SINGLE_CORE_DATAFRAME_CREATION:
            n_cores = 1  # creating dataframes with multiple cores takes much longer...
        else:
            n_cores = self.N_CORES

        chunkSize = min(self.CHUNK_SIZE, maxIndex / n_cores)  # ensure minimum one chunk per core.
        numOfPartitions = int(maxIndex / chunkSize) + 1
        daList = []
        self.ddColNames = []
        for i in range(0, numOfPartitions):
            indexFrom = int(i * chunkSize)
            indexTo = int(min(indexFrom + chunkSize, maxIndex))
            result = dask.delayed(self.createDataframePerElectronRange)(indexFrom, indexTo)
            daList.append(result)
        # self.dd = self.createDataframePerElectronRange(0, maxIndex)

        # Create the electron-indexed dataframe
        self.daListResult = dask.compute(*daList)

        a = np.concatenate(self.daListResult, axis=1)
        da = dask.array.from_array(a.T, chunks=self.CHUNK_SIZE)

        # cols = ('dldPosX', 'dldPosY', 'dldTime', 'delayStage', 'bam', 'dldMicrobunchId', 'dldDetectorId', 'dldSectorId', 'bunchCharge',
        #         'opticalDiode', 'gmdTunnel', 'gmdBda', 'i0Monitor', 'pumpPol', 'timeStamp', 'macroBunchPulseId')
        #
        # cols = tuple(x for x in cols if x in self.daqAddresses)

        cols = self.ddColNames

        self.dd = dask.dataframe.from_array(da, columns=cols)
        # needed as negative values are used to mark bad data
        self.dd = self.dd[self.dd['dldMicrobunchId'] > -1]
        # I propose leaving it like this, since energy calibration depends on microscope parameters and photon energy;
        # CHANGED: default is as before, but if attribute TOF_IN_NS is set to true, it leaves the delay in steps.
        # Changed again. there is aliasing if binning in ns, physical binning needs to happen in tof_steps
        # if self.TOF_IN_NS:
        #     self.dd['dldTime'] = self.dd['dldTime'] * self.TOF_STEP_TO_NS

    def createDataframePerMicrobunch(self):
        """ Create a dataframe indexed by the microbunch ID. The method needs no input parameters.
        """

        if _VERBOSE:
            print('creating microbunch dataframe...')

        arrayCols = []
        ddMicrobunchesColnames = []
        numOfMicrobunches = self.bam.shape[1]
        lengthToPad = numOfMicrobunches - self.opticalDiode.shape[1]

        if 'delayStage' in self.daqAddresses:
            delayStageArray = np.zeros_like(self.bam)
            delayStageArray[:, :] = (self.delayStage[:])[:, None]
            daDelayStage = dask.array.from_array(delayStageArray.flatten(), chunks=self.CHUNK_SIZE)
            arrayCols.append(daDelayStage)
            ddMicrobunchesColnames.append('delayStage')

        if 'streakCam' in self.daqAddresses:
            streakCamArray = np.zeros_like(self.bam)
            streakCamArray[:, :] = (self.streakCam[:, 0])[:, None]
            daStreakCamArray = dask.array.from_array(streakCamArray.flatten(), chunks=self.CHUNK_SIZE)
            arrayCols.append(daStreakCamArray)
            ddMicrobunchesColnames.append('streakCamera')

        if 'bam' in self.daqAddresses:
            daBam = dask.array.from_array(self.bam.flatten(), chunks=(self.CHUNK_SIZE))
            arrayCols.append(daBam)
            ddMicrobunchesColnames.append('bam')

        # if 'dldAux0' in self.daqAddresses:
        #     dldAux0 = self.dldAux0[:, 0]
        #     aux0 = np.ones(self.bam.shape) * dldAux0[:, None]
        #     daAux0 = dask.array.from_array(aux0.flatten(), chunks=(self.CHUNK_SIZE))
        #     arrayCols.append(daAux0)
        #     ddMicrobunchesColnames.append('delayStage')
        #
        # if 'dldAux1' in self.daqAddresses:
        #     dldAux1 = self.dldAux1[:, 1]
        #     aux1 = np.ones(self.bam.shape) * dldAux1[:, None]
        #     daAux1 = dask.array.from_array(aux1.flatten(), chunks=(self.CHUNK_SIZE))
        #     arrayCols.append(daAux1)
        #     ddMicrobunchesColnames.append('delayStage')

        if 'bunchCharge' in self.daqAddresses:
            daBunchCharge = dask.array.from_array(self.bunchCharge[:, 0:numOfMicrobunches].flatten(),
                                                  chunks=(self.CHUNK_SIZE))
            arrayCols.append(daBunchCharge)
            ddMicrobunchesColnames.append('bunchCharge')

        if 'opticalDiode' in self.daqAddresses:
            try:
                paddedOpticalDiode = np.pad(self.opticalDiode, ((0, 0), (0, lengthToPad)), 'constant',
                                            constant_values=(0, 0))
                daOpticalDiode = dask.array.from_array(paddedOpticalDiode.flatten(), chunks=self.CHUNK_SIZE)
            except:
                print('fix optical diode DAQ: Length: ' + str(self.opticalDiode.shape[1]))
                daOpticalDiode = dask.array.from_array(self.opticalDiode[:, 0:numOfMicrobunches].flatten(),
                                                       chunks=self.CHUNK_SIZE)
            arrayCols.append(daOpticalDiode)
            ddMicrobunchesColnames.append('opticalDiode')

        if 'i0Monitor' in self.daqAddresses:
            def I0BunchTrain2Pulses():
                """
                Cut, reshape and average the ADC signal of the I0 monitor. 
                Preliminary procedure: migth be optimized and tunend in to the DAQ 
                """
                i0Cutted = np.array(np.split(-self.i0Monitor[:, 967:54967], 500, axis=1))
                i0BG = i0Cutted[:, :, :39].mean(2)
                return np.nansum(i0Cutted[:, :, 39:90] - i0BG[:, :, None], 2).T

            i0Data = I0BunchTrain2Pulses()
            try:
                paddedI0 = np.pad(i0Data, ((0, 0), (0, numOfMicrobunches - i0Data.shape[1])), 'constant',
                                  constant_values=(0, 0))
                daI0 = dask.array.from_array(paddedI0.flatten(), chunks=self.CHUNK_SIZE)
            except:
                print('fix i0 Monitor DAQ: Length: ' + str(i0Data.shape[1]))
                daI0 = dask.array.from_array(i0Data[:, 0:numOfMicrobunches].flatten(),
                                             chunks=self.CHUNK_SIZE)
            arrayCols.append(daI0)
            ddMicrobunchesColnames.append('i0Monitor')

        if 'pumpPol' in self.daqAddresses:
            pumpPolArray = np.zeros_like(self.bam)
            pumpPolArray[:] = (self.pumpPol[:, 0])[:, None]
            daPumpPol = dask.array.from_array(pumpPolArray.flatten(), chunks=self.CHUNK_SIZE)
            arrayCols.append(daPumpPol)
            ddMicrobunchesColnames.append('pumpPol')

        if 'macroBunchPulseId' in self.daqAddresses:
            macroBunchPulseIdArray = np.zeros_like(self.bam)
            macroBunchPulseIdArray[:, :] = (self.macroBunchPulseId[:, 0])[:, None]
            daMacroBunchPulseId = dask.array.from_array(macroBunchPulseIdArray.flatten(), chunks=(self.CHUNK_SIZE))
            arrayCols.append(daMacroBunchPulseId)
            ddMicrobunchesColnames.append('macroBunchPulseId')

        if 'timeStamp' in self.daqAddresses:
            timeStampArray = np.zeros_like(self.bam)
            timeStampArray[:, :] = (self.timeStamp[:, 0])[:, None]
            daTimeStamp = dask.array.from_array(timeStampArray.flatten(), chunks=(self.CHUNK_SIZE))
            arrayCols.append(daTimeStamp)
            ddMicrobunchesColnames.append('timeStamp')

        da = dask.array.stack(arrayCols)

        # Create the microbunch-indexed dataframe
        # cols = (
        #     'delayStage', 'bam', 'dldAux0', 'dldAux1', 'bunchCharge', 'opticalDiode',  'i0Monitor', 'pumpPol', 'macroBunchPulseId', 'timeStamp')
        # cols = tuple(x for x in cols if x in self.daqAddresses)
        cols = ddMicrobunchesColnames
        self.ddMicrobunches = dask.dataframe.from_array(da.T, columns=cols)

    def storeDataframes(self, fileName=None, path=None, format='parquet', append=False,compression="UNCOMPRESSED"):
        """ Save imported dask dataframe as a parquet or hdf5 file.

        **Parameters**\n
        fileName: str | None
            The file namestring.
        path: str | None (default to ``self.DATA_PARQUET_DIR`` or ``self.DATA_H5_DIR``)
            The path to the folder to save the format-converted data.
        format: str | 'parquet'
            The output file format, possible choices are 'parquet', 'h5' or 'hdf5'.
        append: bool | False (disable data appending as default)
            When using parquet file, allows to append the data to pre-existing files.
        """

        format = format.lower()
        assert format in ['parquet', 'h5', 'hdf5'], 'Invalid format for data input. Please select between parquet or h5'

        # Update instance attributes based on input parameters
        if path is None:
            if format == 'parquet':
                path = self.DATA_PARQUET_DIR
            elif format in ['hdf5', 'h5']:
                path = self.DATA_H5_DIR
        if not os.path.isdir(path):
            raise NotADirectoryError(f'Could not find directory {path}')

        if fileName is None:
            if self.runNumber is None:
                fileName = 'mb{}to{}'.format(self.pulseIdInterval[0], self.pulseIdInterval[-1])
            else:
                fileName = 'run{}'.format(self.runNumber)

        fileName = path + fileName  # TODO: test if naming is correct

        if format == 'parquet':
            if append:
                print(f'Appending data to existing parquet container {fileName}')
            else:
                print(f'Creating parquet container {fileName}')

            with ProgressBar():
                if 'runNumber' not in self.dd.columns: # add run number as column to the dataframe
                    if self.runNumber:
                        run = self.runNumber
                    else:
                        try:
                            run = self.metadata['runInfo']['runNumber']
                            if isinstance(run,list):
                                run = None # in case this is already an agglomerate of runs... useless to add run number
                        except KeyError:
                            run = None
                    if run is not None:
                        self.dd['runNumber'] = run
                        self.ddMicrobunches['runNumber'] = run
                if 'pulseId' not in self.dd.columns: 
                    # add the pulseId (macrobunchId) number as column to the dataframe
                    # macroBunchPulseId is there already, but its reset to 0 at the firsto one...
                    if self.pulseIdInterval:
                        pulseId = self.pulseIdInterval[0]
                    else:
                        try:
                            pulseId = self.metadata['runInfo']['pulseIdInterval'][0]
                            if isinstance(pulseId,list):
                                pulseId = None # in case this is already an agglomerate of runs... useless to add run number
                        except KeyError:
                            pulseId = None
                    if pulseId is not None:
                        self.dd['pulseId'] = self.dd['macroBunchPulseId'] + pulseId
                        self.ddMicrobunches['pulseId'] = self.ddMicrobunches['macroBunchPulseId'] + pulseId

                                                
                self.dd.to_parquet(fileName + "_el", compression=compression, \
                                   append=append, ignore_divisions=True)
                self.ddMicrobunches.to_parquet(fileName + "_mb", compression=compression, \
                                               append=append, ignore_divisions=True)
            print('Saving metadata...')
            try:
                if not bool(self.metadata):
                    self.metadata = self.update_metadata()

                if not append:
                    meta = self.metadata
                else:
                    with open(os.path.join(fileName + '_el', 'run_metadata.txt'),'r') as json_file:
                        old = json.load(json_file)
                    new = self.metadata

                    if 'part_000' not in old.keys():
                        meta={'runInfo':old['runInfo']}
                        meta[f'part_000'] = old
                    else:
                        meta = old
                    
                    last_part_number = max([int(s[5:]) for s in meta.keys() if 'part_' in s])
                    meta[f"part_{last_part_number + 1:03d}"] = new

                    i = meta['runInfo']
                    n = new['runInfo']
                    for key in n.keys():
                        if key == 'runNumber':
                            if not isinstance(i[key],list):
                                i[key] = [i[key],n[key]]
                            else:
                                i[key].append(n[key])

                        elif key == 'pulseIdInterval':
                            if not isinstance(i[key][0],list):
                                i[key] = [i[key],n[key]]
                            else:
                                i[key].append(n[key])
                    
                        elif key == 'timeStart':
                                i[key] = min(datetime.strptime(i[key], '%Y-%m-%d %H:%M:%S'),datetime.strptime(n[key], '%Y-%m-%d %H:%M:%S')).__str__()
                        
                        elif key == 'timeStop':
                                i[key] = max(datetime.strptime(i[key], '%Y-%m-%d %H:%M:%S'),datetime.strptime(n[key], '%Y-%m-%d %H:%M:%S')).__str__()

                        elif key in ['timeDuration', 'electronsPerMacrobunch']:
                            pass # skip those which need recalculation based on the summed values
                        else:
                            try:
                                i[key] += n[key]
                            except TypeError as E:
                                print(E, key)
                        i['timeDuration'] = str(timedelta(seconds=i['timestampDuration']))
                        i['electronsPerMacrobunch'] = i['numberOfElectrons']/i['numberOfMacrobunches']


                with open(os.path.join(fileName + '_el', 'run_metadata.txt'), 'w') as json_file:
                    json.dump(meta, json_file, indent=4)
            except AttributeError as E:
                print(f'failed saving metadata: {E}')

        elif format in ['hdf5', 'h5']:
            if os.path.isdir(fileName):
                print(f'Appending data to existing h5 container {fileName}')
            else:
                print(f'Creating h5 container {fileName}')
            dask.dataframe.to_hdf(self.dd, fileName, '/electrons')
            dask.dataframe.to_hdf(self.ddMicrobunches, fileName, '/microbunches')
        print('saving complete!')

    def getIds(self, runNumber=None, path=None):
        """ Returns the first and the last macrobunch IDs of a given run number.

        **Parameters**\n
        runNumber: int | None (default to ``self.runNumber``)
            The run number from which to read the macrobunch ID interval.
        path: str | None (default to ``self.DATA_RAW_DIR``)
            The path to location where raw HDF5 files are stored.
        
        **Return**\n
        pulseIdInterval: (int, int)
            The macrobunch ID range for a given run number.
        """

        if runNumber is None:
            runNumber = self.runNumber
        else:
            self.runNumber = runNumber

        if path is None:
            path = self.DATA_RAW_DIR

        if not hasattr(self, 'path_to_run'):
            self.path_to_run = misc.get_path_to_run(runNumber, path)

        try:
            from camp.pah.beamtimedaqaccess import H5FileDataAccess, H5FileManager
        except:
            # Gets paths from settings file.
            # Checks for SETTINGS.ini in processor folder.
            # If not there, checks parent directory
            settings = ConfigParser()
            if os.path.isfile(os.path.join(os.path.dirname(__file__), 'SETTINGS.ini')):
                settings.read(os.path.join(os.path.dirname(__file__), 'SETTINGS.ini'))
            else:
                settings.read(os.path.join(os.path.dirname(os.path.dirname(__file__)), 'SETTINGS.ini'))
            path = settings['paths']['DATA_RAW_DIR']

            # needs to import stuff from PAH modules
            import sys
            sys.path.append(settings['paths']['PAH_MODULE_DIR'])
            from camp.pah.beamtimedaqaccess import H5FileDataAccess, H5FileManager

        fileAccess = H5FileDataAccess(H5FileManager(self.path_to_run))
        pulseIdInterval = fileAccess.availablePulseIdInterval(runNumber)

        return pulseIdInterval

    # ==================
    # DEPRECATED METHODS
    # ==================

    def readRun(self, runNumber=None, path=None):
        """ **[DEPRECATED]** Read a run. Generates dd and dd_micrubunches attributes
        as pd.DataFrame containing data from the given run.

        **Parameters**\n
        runNumber: int | None
            number corresponding to the rung to read data from. if None, it uses the value
            defined in the runNumber attribute.
        path: str | None
            path to location where raw HDF5 files are stored. If None, it uses the value from SETTINGS.ini.

        **Raise**\n
            Throws AttributeError if the run number is not given

        :Example:
        ::
        
            processor = DldFlashProcessor()
            processor.readRun(19059)

        """

        print('WARNING: readRun method is obsolete. Please use readData(runNumber=xxx).')

        if path is None:  # allow for using the default path, which can be redefined as class variable.
            path = self.DATA_RAW_DIR
        if runNumber is None:
            runNumber = self.runNumber
            assert runNumber is not None, 'No run number assigned!'
        else:
            self.runNumber = runNumber
        if self.runNumber is None:
            raise AttributeError('Run number not defined. ')
        # Import the dataset
        dldPosXName = "/uncategorised/FLASH1_USER2/FLASH.FEL/HEXTOF.DAQ/DLD1:0/dset"
        dldPosYName = "/uncategorised/FLASH1_USER2/FLASH.FEL/HEXTOF.DAQ/DLD1:1/dset"
        dldTimeName = "/uncategorised/FLASH1_USER2/FLASH.FEL/HEXTOF.DAQ/DLD1:3/dset"

        dldMicrobunchIdName = "/uncategorised/FLASH1_USER2/FLASH.FEL/HEXTOF.DAQ/DLD1:2/dset"
        dldAuxName = "/uncategorised/FLASH1_USER2/FLASH.FEL/HEXTOF.DAQ/DLD1:4/dset"
        # delayStageName = "/Experiment/Pump probe laser/laser delay"
        # ENC.DELAY seems to be the wrong channel! Values appear in groups of exactly the same value
        # delayStageName = "/Experiment/Pump probe laser/delay line IK220.0/ENC.DELAY"
        # Proper channel is column with index 1 of ENC
        delayStageName = "/Experiment/Pump probe laser/delay line IK220.0/ENC"

        bamName = '/Electron Diagnostic/BAM/4DBC3/electron bunch arrival time (low charge)'
        bunchChargeName = '/Electron Diagnostic/Bunch charge/after undulator'
        macroBunchPulseIdName = '/Timing/Bunch train info/index 1.sts'
        opticalDiodeName = '/Experiment/PG/SIS8300 100MHz ADC/CH9/pulse energy/TD'
        gmdTunnelName = '/Photon Diagnostic/GMD/Pulse resolved energy/energy tunnel'
        gmdBdaName = '/Photon Diagnostic/GMD/Pulse resolved energy/energy BDA'

        # adc1Name = '/Experiment/PG/SIS8300 100MHz ADC/CH6/TD'
        # adc2Name = '/Experiment/PG/SIS8300 100MHz ADC/CH7/TD'

        daqAccess = BeamtimeDaqAccess.create(path)

        print('reading DAQ data')
        # ~ print("reading dldPosX")
        self.dldPosX, otherStuff = daqAccess.allValuesOfRun(dldPosXName, runNumber)
        print('run contains macrobunchID from {0:,} to {1:,} \n-> {2:,} total macrobunches'.format(otherStuff[0],
                                                                                                   otherStuff[1],
                                                                                                   otherStuff[1] -
                                                                                                   otherStuff[0]))
        # ~ print("reading dldPosY")
        self.dldPosY, otherStuff = daqAccess.allValuesOfRun(dldPosYName, runNumber)
        # ~ print("reading dldTime")
        self.dldTime, otherStuff = daqAccess.allValuesOfRun(dldTimeName, runNumber)
        # ~ print("reading dldMicrobunchId")
        self.dldMicrobunchId, otherStuff = daqAccess.allValuesOfRun(dldMicrobunchIdName, runNumber)
        # ~ print("reading dldAux")
        self.dldAux, otherStuff = daqAccess.allValuesOfRun(dldAuxName, runNumber)

        # ~ print("reading delayStage")
        self.delayStage, otherStuff = daqAccess.allValuesOfRun(delayStageName, runNumber)
        self.delayStage = self.delayStage[:, 1]

        # ~ print("reading BAM")
        self.bam, otherStuff = daqAccess.allValuesOfRun(bamName, runNumber)
        self.opticalDiode, otherStuff = daqAccess.allValuesOfRun(opticalDiodeName, runNumber)
        # ~ print("reading bunchCharge")
        self.bunchCharge, otherStuff = daqAccess.allValuesOfRun(bunchChargeName, runNumber)
        self.macroBunchPulseId, otherStuff = daqAccess.allValuesOfRun(macroBunchPulseIdName, runNumber)
        self.macroBunchPulseId -= otherStuff[0]
        self.gmdTunnel, otherStuff = daqAccess.allValuesOfRun(gmdTunnelName, runNumber)
        self.gmdBda, otherStuff = daqAccess.allValuesOfRun(gmdBdaName, runNumber)
        electronsToCount = self.dldPosX.copy().flatten()
        electronsToCount = np.nan_to_num(electronsToCount)
        electronsToCount = electronsToCount[electronsToCount > 0]
        electronsToCount = electronsToCount[electronsToCount < 10000]
        numOfElectrons = len(electronsToCount)
        print("Number of electrons: {0:,} ".format(numOfElectrons))
        print("Creating data frame: Please wait...")
        self.createDataframePerElectron()
        self.createDataframePerMicrobunch()
        print('dataframe created')

    def readInterval(self, pulseIdInterval, path=None):
        """ **[DEPRECATED]** Access to data by an macrobunch ID interval.
        Useful for scans that would otherwise hit the machine's memory limit.

        **Parameters**\n
        pulseIdInterval: (int, int)
            The starting and ending macrobunch IDs, [start, end).
        path: str | None (default to ``self.DATA_RAW_DIR``)
            The path to location where raw HDF5 files are stored.
        """

        # allow for using the default path, which can be redefined as class variable. leaving retrocompatibility
        print('WARNING: readInterval method is obsolete. Please use readData(pulseIdInterval=xxx).')

        if path is None:
            path = self.DATA_RAW_DIR

        self.pulseIdInterval = pulseIdInterval
        # Import the dataset
        dldPosXName = "/uncategorised/FLASH1_USER2/FLASH.FEL/HEXTOF.DAQ/DLD1:0/dset"
        dldPosYName = "/uncategorised/FLASH1_USER2/FLASH.FEL/HEXTOF.DAQ/DLD1:1/dset"
        dldTimeName = "/uncategorised/FLASH1_USER2/FLASH.FEL/HEXTOF.DAQ/DLD1:3/dset"

        dldMicrobunchIdName = "/uncategorised/FLASH1_USER2/FLASH.FEL/HEXTOF.DAQ/DLD1:2/dset"
        dldAuxName = "/uncategorised/FLASH1_USER2/FLASH.FEL/HEXTOF.DAQ/DLD1:4/dset"
        # delayStageName = "/Experiment/Pump probe laser/laser delay"
        # ENC.DELAY seems to be the wrong channel! Values appear in groups of ~10 identical values
        # -> ENC.DELAY is read out with 1 Hz
        # delayStageName = "/Experiment/Pump probe laser/delay line IK220.0/ENC.DELAY"
        # Proper channel is culumn with index 1 of ENC
        delayStageName = "/Experiment/Pump probe laser/delay line IK220.0/ENC"

        bamName = '/Electron Diagnostic/BAM/4DBC3/electron bunch arrival time (low charge)'
        bunchChargeName = '/Electron Diagnostic/Bunch charge/after undulator'
        macroBunchPulseIdName = '/Timing/Bunch train info/index 1.sts'
        opticalDiodeName = '/Experiment/PG/SIS8300 100MHz ADC/CH9/pulse energy/TD'
        gmdTunnelName = '/Photon Diagnostic/GMD/Pulse resolved energy/energy tunnel'
        gmdBdaName = '/Photon Diagnostic/GMD/Pulse resolved energy/energy BDA'

        # adc1Name = '/Experiment/PG/SIS8300 100MHz ADC/CH6/TD'
        # adc2Name = '/Experiment/PG/SIS8300 100MHz ADC/CH7/TD'

        daqAccess = BeamtimeDaqAccess.create(path)

        print('reading DAQ data')
        # ~ print("reading dldPosX")
        self.dldPosX = daqAccess.valuesOfInterval(dldPosXName, pulseIdInterval)
        # ~ print("reading dldPosY")
        self.dldPosY = daqAccess.valuesOfInterval(dldPosYName, pulseIdInterval)
        # ~ print("reading dldTime")
        self.dldTime = daqAccess.valuesOfInterval(dldTimeName, pulseIdInterval)
        # ~ print("reading dldMicrobunchId")
        self.dldMicrobunchId = daqAccess.valuesOfInterval(dldMicrobunchIdName, pulseIdInterval)
        # ~ print("reading dldAux")
        self.dldAux = daqAccess.valuesOfInterval(dldAuxName, pulseIdInterval)

        # ~ print("reading delayStage")
        self.delayStage = daqAccess.valuesOfInterval(delayStageName, pulseIdInterval)
        self.delayStage = self.delayStage[:, 1]

        # ~ print("reading BAM")
        self.bam = daqAccess.valuesOfInterval(bamName, pulseIdInterval)
        self.opticalDiode = daqAccess.valuesOfInterval(opticalDiodeName, pulseIdInterval)
        # ~ print("reading bunchCharge")
        self.bunchCharge = daqAccess.valuesOfInterval(bunchChargeName, pulseIdInterval)
        self.macroBunchPulseId = daqAccess.valuesOfInterval(macroBunchPulseIdName, pulseIdInterval)
        # self.macroBunchPulseId -= self.macroBunchPulseId[self.macroBunchPulseId > 0].min()
        self.macroBunchPulseId -= pulseIdInterval[0]
        self.gmdTunnel = daqAccess.valuesOfInterval(gmdTunnelName, pulseIdInterval)
        self.gmdBda = daqAccess.valuesOfInterval(gmdBdaName, pulseIdInterval)
        electronsToCount = self.dldPosX.copy().flatten()
        electronsToCount = np.nan_to_num(electronsToCount)
        electronsToCount = electronsToCount[electronsToCount > 0]
        electronsToCount = electronsToCount[electronsToCount < 10000]
        numOfElectrons = len(electronsToCount)
        print("Number of electrons: {0:,} ".format(numOfElectrons))
        print("Creating data frame: Please wait...")
        self.createDataframePerElectron()
        self.createDataframePerMicrobunch()
        print('dataframe created')



# -*- coding: utf-8 -*-
"""

@author: Davide Curcio
"""
from distutils.core import setup, Extension

XPSdoniachs = Extension('XPSdoniachs_ext',
                        sources=['XPSdoniachs_ext.cpp'],
                        libraries=["boost_python37"])

setup(name='XPSdoniachs',
      packages=['.'],
      ext_modules=[XPSdoniachs])



from processor.DldProcessor import DldProcessor
import sys, os
import glob
import json
import h5py
import numpy as np
from pandas import Series, DataFrame, concat, MultiIndex
from dask import delayed, compute
import dask.dataframe as dd
from dask.diagnostics import ProgressBar
from pathlib import Path
from functools import reduce
from configparser import ConfigParser
from multiprocessing import Pool, cpu_count
from itertools import compress

"""
    author:: Muhammad Zain Sohail, Steinn Ymir Agustsson
    Some functions taken or adapted from
    https://gitlab.desy.de/christopher.passow/flash-daq-hdf/
"""

# assert sys.version_info >= (3,8), f"Requires at least Python version 3.8,\
#     but was {sys.version}"

class DldFlashProcessorExpress(DldProcessor):
    """  
    The class generates multiindexed multidimensional pandas dataframes 
    from the new FLASH dataformat resolved by both macro and microbunches alongside electrons.
    """
    def __init__(self, runNumber = None, channels = None, settings = None, beamtime_dir=None, 
                parquet_path=None, parquet_dir=None, beamtime_id=None, year=None, daq="fl1user2", silent=False):
        super().__init__(settings=settings,silent=silent)
        self.runNumber = runNumber
        
        # Parses and stores the channel names that are defined by the json file
        if channels is None:
            all_channel_list_dir = os.path.join(Path(__file__).parent.absolute(), "channels.json")
        else:
            all_channel_list_dir = channels
        # Read all channel info from a json file
        with open(all_channel_list_dir, "r") as json_file:
            self.all_channels = json.load(json_file)
        self.channels = self.availableChannels # Set all channels, exluding pulseId as default
        
        # Prases to locate the raw beamtime directory from settings file 
        if 'paths' in self.settings.keys():
            self.DATA_RAW_DIR = Path(self.settings['paths']['data_raw_dir'])
        else:    
            if beamtime_dir is None:
                if beamtime_id is None:
                    beamtime_id = self.settings['processor']['beamtime_id']
                if year is None:
                    year = self.settings['processor']['year']
                if beamtime_id is None or year is None:
                    raise ValueError('Either the beamtime_dir or beamtime_id and year or a settings file\
                    containing such info is needed to find the data.')
                self.beamtime_dir = Path(f'/asap3/flash/gpfs/pg2/{year}/data/{beamtime_id}/')

            else:
                self.beamtime_dir = Path(beamtime_dir)

            # Folder naming convention till end of October
            self.DATA_RAW_DIR = self.beamtime_dir.joinpath('raw/hdf/express')
            # Use new convention if express doesn't exist
            if not self.DATA_RAW_DIR.exists(): 
                self.DATA_RAW_DIR = self.beamtime_dir.joinpath(f'raw/hdf/{daq.upper()}')
        
        # Parses and creates directory to store parquet files
        if 'paths' in self.settings.keys():
            self.DATA_PARQUET_DIR = Path(self.settings['paths']['data_parquet_dir'])
        else:
            if parquet_dir is None:
                if parquet_path is None:
                    self.parquet_path = 'processed/parquet'
                self.DATA_PARQUET_DIR = self.beamtime_dir.joinpath(self.parquet_path)
            else:
                self.DATA_PARQUET_DIR = Path(parquet_dir)

            if not self.DATA_PARQUET_DIR.exists():
                os.mkdir(self.DATA_PARQUET_DIR)    
            
    @property
    def availableChannels(self):
        """Returns the channel names that are available for use, 
        excluding pulseId, defined by the json file"""
        available_channels = list(self.all_channels.keys())
        available_channels.remove('pulseId')
        return available_channels
    
    @property
    def channelsPerPulse(self):
        """Returns a list of channels with per_pulse format, 
        including all auxillary channels"""
        channels_per_pulse = []
        for key in self.availableChannels:
            if self.all_channels[key]['format'] == 'per_pulse':
                if key == 'dldAux':
                    for aux_key in self.all_channels[key]['dldAuxChannels'].keys():
                        channels_per_pulse.append(aux_key)
                else:    
                    channels_per_pulse.append(key)
        return channels_per_pulse
    
    @property
    def channelsPerElectron(self):
        """Returns a list of channels with per_electron format"""
        return [key for key in self.availableChannels 
                if self.all_channels[key]['format'] == 'per_electron']
    
    @property
    def channelsPerTrain(self):
        """Returns a list of channels with per_train format"""
        return [key for key in self.availableChannels 
                if self.all_channels[key]['format'] == 'per_train']

    def resetMultiIndex(self):
        """Resets the index per pulse and electron"""
        self.index_per_electron = None
        self.index_per_pulse = None

    def createMultiIndexPerElectron(self, h5_file):
        """Creates an index per electron using pulseId 
        for usage with the electron resolved pandas dataframe"""

        # Macrobunch IDs obtained from the pulseId channel
        [train_id, np_array] = self.createNumpyArrayPerChannel(h5_file, "pulseId")

        # Create a series with the macrobunches as index and microbunches as values
        macrobunches = Series((np_array[i] for i in train_id.index),
            name="pulseId",
            index=train_id) - self.UBID_OFFSET

        # Explode dataframe to get all microbunch vales per macrobunch,
        # remove NaN values and convert to type int
        microbunches = macrobunches.explode().dropna().astype(int)
        
        # Create temporary index values 
        index_temp = MultiIndex.from_arrays((microbunches.index, microbunches.values),
                    names=["trainId", "pulseId"])

        # Calculate the electron counts per pulseId
        # unique preserves the order of appearance
        electron_counts = index_temp.value_counts()[index_temp.unique()].values

        # Series object for indexing with electrons
        electrons = Series([np.arange(electron_counts[i]) 
                            for i in range(electron_counts.size)]).explode()

        # Create a pandas multiindex using the exploded datasets
        self.index_per_electron = MultiIndex.from_arrays(
            (microbunches.index, microbunches.values, electrons),
            names=["trainId", "pulseId", 'electronId'])

    def createMultiIndexPerPulse(self, train_id, np_array):
        """Creates an index per pulse using a pulse resovled channel's 
        macrobunch ID, for usage with the pulse resolved pandas dataframe"""

        # Create a pandas multiindex, useful to compare electron and 
        # pulse resolved dataframes
        self.index_per_pulse = MultiIndex.from_product(
            (train_id, np.arange(0, np_array.shape[1])),
            names=["trainId", "pulseId"])

    def createNumpyArrayPerChannel(self, h5_file, channel):
        """Returns a numpy Array for a given channel name for a given file"""
        # Get the data from the necessary h5 file and channel
        group = h5_file[self.all_channels[channel]["group_name"]]
        channel_dict = self.all_channels[channel]  # channel parameters

        train_id = Series(group["index"], name="trainId")  # macrobunch
        # unpacks the timeStamp or value
        if channel == "timeStamp":
            np_array = group["time"][()]
        else:
            np_array = group["value"][()]

        # Uses predefined axis and slice from the json file
        # to choose correct dimension for necessary channel
        if "axis" in channel_dict:
            np_array = np.take(
                np_array, channel_dict["slice"], axis=channel_dict["axis"])
        return train_id, np_array

    def createDataframePerChannel(self, h5_file, channel):
        """Returns a pandas DataFrame for a given channel name for
        a given file. The Dataframe contains the MultiIndex and returns 
        depending on the channel's format"""
        [train_id, np_array] = self.createNumpyArrayPerChannel(
            h5_file, channel)  # numpy Array created
        channel_dict = self.all_channels[channel]  # channel parameters
        
        # If np_array is size zero, fill with NaNs
        if np_array.size == 0:
            np_array = np.full_like(train_id, np.nan, dtype = np.double)
            return Series((np_array[i] for i in train_id.index), name = channel, 
                          index = train_id)
                    
        # Electron resolved data is treated here
        if channel_dict["format"] == "per_electron":
            # Creates the index_per_electron if it does not exist for a given file
            if self.index_per_electron is None:
                self.createMultiIndexPerElectron(h5_file)

            # The microbunch resolved data is exploded and converted to dataframe, 
            # afterwhich the MultiIndex is set 
            # The NaN values are dropped, alongside the pulseId = 0 (meaningless)
            return (Series((np_array[i] for i in train_id.index), name=channel)
                .explode()
                .dropna()
                .to_frame()
                .set_index(self.index_per_electron)
                .drop(index = np.arange(-self.UBID_OFFSET, 0), level = 1, errors = 'ignore'))
        
        
        # Pulse resolved data is treated here
        elif channel_dict["format"] == "per_pulse":
            # Special case for auxillary channels which checks the channel dictionary 
            # for correct slices and creates a multicolumn pandas dataframe
            if channel == "dldAux":
                # The macrobunch resolved data is repeated 499 times to be
                # comapred to electron resolved data for each auxillary channel
                # and converted to a multicolumn dataframe     
                data_frames = (
                    Series((np_array[i, value] for i in train_id.index),
                    name=key,
                    index=train_id)
                    .to_frame()
                    for key, value in channel_dict["dldAuxChannels"].items())

                # Multiindex set and combined dataframe returned
                return reduce(DataFrame.combine_first, data_frames)
            
            else:
                # For all other pulse resolved channels, macrobunch resolved 
                # data is exploded to a dataframe and the MultiIndex set
            
                # Creates the index_per_pulse for the given channel
                self.createMultiIndexPerPulse(train_id, np_array)
                return (Series((np_array[i] for i in train_id.index), name=channel)
                    .explode()
                    .to_frame()
                    .set_index(self.index_per_pulse))

        elif channel_dict["format"] == "per_train":
                return (Series((np_array[i] for i in train_id.index), name=channel)
                    .to_frame()
                    .set_index(train_id))
        
        else:
            raise ValueError(channel + "has an undefined format. Available formats are \
                per_pulse, per_electron and per_train")

    def concatenateChannels(self, h5_file, format_=None):
        """Returns a concatenated pandas DataFrame for either all pulse
        resolved or all electron resolved channels."""
        valid_names = [each_name
            for each_name in self.channels
            if each_name in self.all_channels]  # filters for valid channels
        # Only channels with the defined format are selected and stored 
        # in an iterable list
        if format_ is not None:
            channels = [each_name
                for each_name in valid_names
                if self.all_channels[each_name]["format"] == format_]
        else:
            channels = [each_name for each_name in valid_names]

        # if the defined format has channels, returns a concatenatd Dataframe.
        # Otherwise returns empty Dataframe.
        if channels:
            data_frames = (
                self.createDataframePerChannel(h5_file, each)
                for each in channels) 
            return reduce(lambda left, right: left.join(right, how='outer'), data_frames)

    def createDataframePerFile(self, file_path):
        """Returns two pandas DataFrames constructed for the given file.
        The DataFrames contains the datasets from the iterable in the
        order opposite to specified by channel names. One DataFrame is 
        pulse resolved and the other electron resolved.
        """
        # Loads h5 file and creates two dataframes
        with h5py.File(file_path, "r") as h5_file:
            self.resetMultiIndex()  # Reset MultiIndexes for next file
            return self.concatenateChannels(h5_file) 
          
    def runFilesNames(self, run_number, daq, raw_data_dir):
        """Returns all filenames of given run located in directory for the given daq."""
        stream_name_prefixes = {"pbd": "GMD_DATA_gmd_data",
                                "pbd2": "FL2PhotDiag_pbd2_gmd_data",
                                "fl1user1": "FLASH1_USER1_stream_2",
                                "fl1user2": "FLASH1_USER2_stream_2",
                                "fl1user3": "FLASH1_USER3_stream_2",
                                "fl2user1": "FLASH2_USER1_stream_2",
                                "fl2user2": "FLASH2_USER2_stream_2"}
        date_time_section = lambda filename: str(filename).split("_")[-1]
        return sorted(Path(raw_data_dir)
            .glob(f"{stream_name_prefixes[daq]}_run{run_number}_*.h5"),
            key=date_time_section)
    
    def h5_to_parquet(self, h5, prq):
        try:
            (self.createDataframePerFile(h5)
            .reset_index(level=['trainId', 'pulseId','electronId'])
            .to_parquet(prq, compression = None, index = False))
        except ValueError as e:
            self.failed_str.append(f'{prq}: {e}')
            self.prq_names.remove(prq)
            
    def fill_na(self):
        """Routine to fill the NaN values with intrafile forward filling. """
        # First use forward filling method to fill each file's pulse resolved channels.
        for i in range(len(self.dfs)):
            self.dfs[i][self.channelsPerPulse] = self.dfs[i][self.channelsPerPulse].fillna(method="ffill")
        
        # This loop forward fills between the consective files. 
        # The first run file will have NaNs, unless another run before it has been defined.
        for i in range(1, len(self.dfs)):
            subset = self.dfs[i][self.channelsPerPulse] # Take only pulse channels
            is_null = subset.loc[0].isnull().values.compute() # Find which column(s) contain NaNs.
            # Statement executed if there is more than one NaN value in the first row from all columns
            if is_null.sum() > 0: 
                # Select channel names with only NaNs
                channels_to_overwrite = list(compress(self.channelsPerPulse, is_null[0]))
                # Get the values for those channels from previous file
                values = self.dfs[i-1][self.channelsPerPulse].tail(1).values[0]
                # Fill all NaNs by those values
                subset[channels_to_overwrite] = subset[channels_to_overwrite].fillna(
                                                                dict(zip(channels_to_overwrite, values)))
                # Overwrite the dataframes with filled dataframes
                self.dfs[i][self.channelsPerPulse] = subset

    def readData(self, runs=None, ignore_missing_runs=False, settings=None, channels=None,
             beamtime_dir=None, parquet_path=None, beamtime_id=None, year=None,
            daq="fl1user2"):
        """ Read express data from DAQ, generating a parquet in between.

        Args:
            runs: int | list
                run number or list of run numbers to load
            ignore_missing_runs: bool
                if False, rises FileNotFoundError in case files for a run are not available.
            settings: str | path
                pointer to the ini settings file, handeled by the dldProcessor class. It can
                be the name of a default settings file found in the settings dir of the repo,
                or the path to a specific settings file.
            channels: list
                list of channel names to include in the dataframe. if none defaults to all available channels
            beamtime_dir: str | path
                path to the raaw data. If none, its inferred from the settings file
            parquet_path: str | path
                path relative to beamtime_dir where to storethe parquet files. Defaults to "beamtime_dir/processed/parquet"
            beamtime_id: int
                the id of the beamtime. If none it is inferred from settings
            year: int
                the year of the beamtime. If none it is inferred from settings
            daq: str
                the daq containig the data. If none it is inferred from settings
                
        returns:
            prc: DldProcessor
                returns an instance of the processor class, with electron and pulse dataframes loaded.
        """
        if (self.runNumber or runs) is None:
            raise ValueError('Must provide a run or list of runs!')
                        
        if not runs:
            runs = self.runNumber 
            
        # create a per_file directory
        self.temp_parquet_dir = self.DATA_PARQUET_DIR.joinpath('per_file')
        if not self.temp_parquet_dir.exists():
            os.mkdir(self.temp_parquet_dir)

        # Prepare a list of names for the files to read and parquets to write
        try: 
            runs_str = f'Runs {runs[0]} to {runs[-1]}'
        except TypeError:
            runs_str = f'Run {runs}'
            runs = [runs]
        parquet_name = f'{self.temp_parquet_dir}/'
        all_files = []
        for run in runs:
            files = self.runFilesNames(run, daq, self.DATA_RAW_DIR)
            [all_files.append(f) for f in files]
            if len(files) == 0 and not ignore_missing_runs:
                raise FileNotFoundError(f'No file found for run {run}')

        self.prq_names = [parquet_name + all_files[i].stem for i in range(len(all_files))]
        missing_files = []
        missing_prq_names = []
        
        # only read and write files which were not read already 
        for i in range(len(self.prq_names)): 
            if not Path(self.prq_names[i]).exists():
                missing_files.append(all_files[i])
                missing_prq_names.append(self.prq_names[i])
        
        print(f'Reading {runs_str}: {len(missing_files)} new files of {len(all_files)} total.')
        self.failed_str = []
        
        self.resetMultiIndex()  # initializes the indices for h5_to_parquet
        
        # Set cores for multiprocessing
        N_CORES = len(missing_files)
        if N_CORES > cpu_count() - 1:
            N_CORES = cpu_count() - 1
        
        # Read missing files using multiple cores
        if len(missing_files) > 0:
            with Pool(processes=N_CORES) as pool:    
                pool.starmap(self.h5_to_parquet, tuple(zip(missing_files, missing_prq_names)))
                            
        if len(self.failed_str)>0:
            print(f'Failed reading {len(self.failed_str)} files of{len(all_files)}:')
            for s in self.failed_str:
                print(f'\t- {s}')
        if len(self.prq_names)==0:
            raise ValueError('No data available. Probably failed reading all h5 files')
        else:
            print(f'Loading {len(self.prq_names)} dataframes. Failed reading {len(all_files)-len(self.prq_names)} files.')  
            self.dfs = [dd.read_parquet(fn) for fn in self.prq_names]
            self.fill_na()
            df = dd.concat(self.dfs)
            df_electron = df.dropna(subset=self.channelsPerElectron)
            pulse_columns = ['trainId','pulseId','electronId'] + self.channelsPerPulse
            df_pulse = df[pulse_columns]
            df_pulse = df_pulse[(df_pulse['electronId']==0)|(np.isnan(df_pulse['electronId']))]

        self.dd  = df_electron.repartition(npartitions=len(self.prq_names))
        self.ddMicrobunches = df_pulse.repartition(npartitions=len(self.prq_names))



# -*- coding: utf-8 -*-

"""
@author: Steinn Ymir Agustsson
"""
import matplotlib.pyplot as plt
import numpy as np
from ipywidgets import widgets, interactive_output

cmaps = ['viridis', 'plasma', 'inferno', 'magma', 'Greys', 'Purples', 'Blues', 'Greens', 'Oranges', 'Reds',
         'YlOrBr', 'YlOrRd', 'OrRd', 'PuRd', 'RdPu', 'BuPu',
         'GnBu', 'PuBu', 'YlGnBu', 'PuBuGn', 'BuGn', 'YlGn', 'binary', 'gist_yarg', 'gist_gray', 'gray', 'bone', 'pink',
         'spring', 'summer', 'autumn', 'winter', 'cool', 'Wistia',
         'hot', 'afmhot', 'gist_heat', 'copper', 'PiYG', 'PRGn', 'BrBG', 'PuOr', 'RdGy', 'RdBu',
         'RdYlBu', 'RdYlGn', 'Spectral', 'coolwarm', 'bwr', 'seismic', 'Pastel1', 'Pastel2', 'Paired', 'Accent',
         'Dark2', 'Set1', 'Set2', 'Set3',
         'tab10', 'tab20', 'tab20b', 'tab20c', 'flag', 'prism', 'ocean', 'gist_earth', 'terrain', 'gist_stern',
         'gnuplot', 'gnuplot2', 'CMRmap', 'cubehelix', 'brg', 'hsv',
         'gist_rainbow', 'rainbow', 'jet', 'nipy_spectral', 'gist_ncar']
    
def isnotebook():
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True  # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter

def orthoslices_3D(data, normalize=True, continuous_update=False, **kwargs):
    if not isnotebook:
        print('Function not suited for working outside of Jupyter notebooks!')
    else:
        get_ipython().magic('matplotlib notebook')
        get_ipython().magic('matplotlib notebook')

    e_range, x_range, y_range = data.shape
    dmin = np.amin(data)
    dmax = np.amax(data)

    w_e = widgets.IntSlider(value=e_range // 2, min=0, max=e_range - 1, step=1, description='Energy:',
                            disabled=False, continuous_update=continuous_update, orientation='horizontal', readout=True,
                            readout_format='d')
    w_kx = widgets.IntSlider(value=x_range // 2, min=0, max=x_range - 1, step=1, description='kx:',
                             disabled=False, continuous_update=continuous_update, orientation='horizontal',
                             readout=True,
                             readout_format='d')
    w_ky = widgets.IntSlider(value=y_range // 2, min=0, max=y_range - 1, step=1, description='ky:', disabled=False,
                             continuous_update=continuous_update, orientation='horizontal', readout=True,
                             readout_format='d')
    w_clim = widgets.FloatRangeSlider(value=[.1, .9], min=0, max=1, step=0.001, description='Contrast:', disabled=False,
                                      continuous_update=continuous_update, orientation='horizontal', readout=True,
                                      readout_format='.1f')
    w_cmap = widgets.Dropdown(options=cmaps, value='terrain', description='colormap:', disabled=False)
    w_bin = widgets.BoundedIntText(value=1, min=1, max=min(data.shape), step=1, description='resample:', disabled=False)
    w_interpolate = widgets.Checkbox(value=True, description='Interpolate', disabled=False)
    w_grid = widgets.Checkbox(value=False, description='Grid', disabled=False)
    w_trackers = widgets.Checkbox(value=True, description='Trackers', disabled=False)
    w_trackercol = widgets.ColorPicker(concise=False, description='tracker line color', value='orange')

    ui_pos = widgets.HBox([w_e, w_kx, w_ky])
    ui_color = widgets.HBox([widgets.VBox([w_clim, w_cmap]),
                             widgets.VBox([w_bin, w_interpolate, w_grid]),
                             widgets.VBox([w_trackers, w_trackercol]),
                             ])

    children = [ui_pos, ui_color]
    tab = widgets.Tab(children=children, )
    tab.set_title(0, 'data select')
    tab.set_title(1, 'colormap')

    figsize = kwargs.pop('figsize', (5, 5))
    fig = plt.figure(figsize=figsize, **kwargs)
    plt.tight_layout()
    # [left, bottom, width, height]
    # fig.locator_params(nbins=4)

    # cbar_ax = fig.add_axes([.05,.4,.05,4], xticklabels=[], yticklabels=[])
    # cbar_ax.yaxis.set_major_locator(plt.LinearLocator(5))

    img_ax = fig.add_axes([.15, .4, .4, .4], xticklabels=[], yticklabels=[])
    img_ax.xaxis.set_major_locator(plt.LinearLocator(5))
    img_ax.yaxis.set_major_locator(plt.LinearLocator(5))

    xproj_ax = fig.add_axes([.15, .1, .4, .28], xticklabels=[], yticklabels=[])
    xproj_ax.set_xlabel('$k_x$')
    xproj_ax.xaxis.set_major_locator(plt.LinearLocator(5))

    yproj_ax = fig.add_axes([.57, .4, .28, .4], xticklabels=[], yticklabels=[])
    yproj_ax.yaxis.set_label_position("right")
    yproj_ax.set_ylabel('$k_y$')
    yproj_ax.yaxis.set_major_locator(plt.LinearLocator(5))

    for ax in [img_ax, yproj_ax, xproj_ax]:  # ,cbar_ax]:
        ax.tick_params(axis="both", direction="in", bottom=True, top=True, left=True, right=True, which='both')

    clim_ = 0.01, .99

    e_img = norm_img(data[data.shape[0] // 2, :, :])
    y_img = norm_img(data[:, data.shape[1] // 2, :])
    x_img = norm_img(data[:, :, data.shape[2] // 2].T)
    e_plot = img_ax.imshow(e_img, cmap='terrain', aspect='auto', interpolation='gaussian',
                           clim=clim_, )  # origin='lower')
    x_plot = yproj_ax.imshow(x_img, cmap='terrain', aspect='auto', interpolation='gaussian',
                             clim=clim_, )  # origin='lower')
    y_plot = xproj_ax.imshow(y_img, cmap='terrain', aspect='auto', interpolation='gaussian',
                             clim=clim_, )  # origin='lower')

    pe_x = img_ax.axvline(x_range / 2, c='orange')
    pe_y = img_ax.axhline(y_range / 2, c='orange')
    px_x = xproj_ax.axvline(x_range / 2, c='orange')
    px_e = xproj_ax.axhline(e_range / 2, c='orange')
    py_y = yproj_ax.axhline(y_range / 2, c='orange')
    py_e = yproj_ax.axvline(e_range / 2, c='orange')

    def update(e, kx, ky, clim, cmap, binning, interpolate, grid, trackers, trackerscol):
        if normalize:
            e_img = norm_img(data[e, :, :][::binning, ::binning])
            y_img = norm_img(data[:, ky, :][::binning, ::binning])
            x_img = norm_img(data[:, :, kx][::binning, ::binning])
        else:
            e_img = data[e, :, :][::binning, ::binning]
            y_img = data[:, ky, :][::binning, ::binning]
            x_img = data[:, :, kx][::binning, ::binning]
        for axis, plot, img in zip([img_ax, yproj_ax, xproj_ax], [e_plot, x_plot, y_plot], [e_img, x_img.T, y_img]):

            plot.set_data(img)
            plot.set_clim(clim)
            plot.set_cmap(cmap)
            axis.grid(grid)
            if interpolate:
                plot.set_interpolation('gaussian')
            else:
                plot.set_interpolation(None)
            if trackers:
                pe_x.set_xdata(kx)
                pe_x.set_color(trackerscol)
                pe_y.set_ydata(ky)
                pe_y.set_color(trackerscol)
                px_x.set_xdata(kx)
                px_x.set_color(trackerscol)
                px_e.set_ydata(e)
                px_e.set_color(trackerscol)
                py_y.set_ydata(ky)
                py_y.set_color(trackerscol)
                py_e.set_xdata(e)
                py_e.set_color(trackerscol)

    interactive_plot = interactive_output(update, {'e': w_e,
                                                   'kx': w_kx,
                                                   'ky': w_ky,
                                                   'clim': w_clim,
                                                   'cmap': w_cmap,
                                                   'binning': w_bin,
                                                   'interpolate': w_interpolate,
                                                   'grid': w_grid,
                                                   'trackers': w_trackers,
                                                   'trackerscol': w_trackercol, });
    display(interactive_plot, tab)
    # display(tab)

    # return fig


def orthoslices_4D(data, axis_order=['E', 'kx', 'ky', 'kz'], normalize=True, continuous_update=True, **kwargs):
    if not isnotebook:
        raise EnvironmentError('Function not suited for working outside of Jupyter notebooks!')
    else:
        get_ipython().magic('matplotlib notebook')
        get_ipython().magic('matplotlib notebook')

    assert len(data.shape) == 4, 'Data should be 4-dimensional, but data has {} dimensions'.format(data.shape)

    # make controls for data slicers
    # slicers = []
    # for shape, name in zip(data.shape, axis_order):
    #     slicers.append(widgets.IntSlider(value=shape // 2,
    #                                      min=0,
    #                                      max=shape - 1,
    #                                      step=1,
    #                                      description=name,
    #                                      disabled=False,
    #                                      continuous_update=False,
    #                                      orientation='horizontal',
    #                                      readout=True,
    #                                      readout_format='d'
    #                                      ))

    e_range, x_range, y_range, z_range = data.shape

    w_e = widgets.IntSlider(value=e_range // 2, min=0, max=e_range - 1, step=1, description='Energy:',
                            disabled=False, continuous_update=continuous_update, orientation='horizontal', readout=True,
                            readout_format='d')
    w_kx = widgets.IntSlider(value=x_range // 2, min=0, max=x_range - 1, step=1, description='kx:',
                             disabled=False, continuous_update=continuous_update, orientation='horizontal',
                             readout=True,
                             readout_format='d')
    w_ky = widgets.IntSlider(value=y_range // 2, min=0, max=y_range - 1, step=1, description='ky:', disabled=False,
                             continuous_update=continuous_update, orientation='horizontal', readout=True,
                             readout_format='d')
    w_kz = widgets.IntSlider(value=z_range // 2, min=0, max=z_range - 1, step=1, description='kz:', disabled=False,
                             continuous_update=continuous_update, orientation='horizontal', readout=True,
                             readout_format='d')

    slicers = [w_e, w_kx, w_ky, w_kz]
    ui_slicers = widgets.HBox(slicers)

    # make controls for graphics appearance
    w_clim = widgets.FloatRangeSlider(value=[.1, .9], min=0, max=1, step=0.001, description='Contrast:', disabled=False,
                                      continuous_update=True, orientation='horizontal', readout=True,
                                      readout_format='.1f')
    w_cmap = widgets.Dropdown(options=cmaps, value='terrain', description='colormap:', disabled=False)
    w_bin = widgets.BoundedIntText(value=1, min=1, max=min(data.shape), step=1, description='resample:', disabled=False)
    w_interpolate = widgets.Checkbox(value=True, description='Interpolate', disabled=False)
    w_grid = widgets.Checkbox(value=False, description='Grid', disabled=False)
    w_trackers = widgets.Checkbox(value=True, description='Trackers', disabled=False)
    w_trackercol = widgets.ColorPicker(concise=False, description='tracker line color', value='orange')
    ui_color = widgets.HBox([widgets.VBox([w_clim, w_cmap]),
                             widgets.VBox([w_bin, w_interpolate, w_grid]),
                             widgets.VBox([w_trackers, w_trackercol]),
                             ])

    tab = widgets.Tab(children=[ui_slicers, ui_color], )
    tab.set_title(0, 'Data slicing')
    tab.set_title(1, 'Graphics')

    figsize = kwargs.pop('figsize', (5, 5))
    fig = plt.figure(figsize=figsize, **kwargs)
    plt.tight_layout()

    img_ax = fig.add_axes([.15, .4, .4, .4], xticklabels=[], yticklabels=[])
    img_ax.xaxis.set_major_locator(plt.LinearLocator(5))
    img_ax.yaxis.set_major_locator(plt.LinearLocator(5))

    xproj_ax = fig.add_axes([.15, .1, .4, .28], xticklabels=[], yticklabels=[])
    xproj_ax.set_xlabel('$k_x$')
    xproj_ax.xaxis.set_major_locator(plt.LinearLocator(5))

    yproj_ax = fig.add_axes([.57, .4, .28, .4], xticklabels=[], yticklabels=[])
    yproj_ax.yaxis.set_label_position("right")
    yproj_ax.set_ylabel('$k_y$')
    yproj_ax.yaxis.set_major_locator(plt.LinearLocator(5))

    for ax in [img_ax, yproj_ax, xproj_ax]:  # ,cbar_ax]:
        ax.tick_params(axis="both", direction="in", bottom=True, top=True, left=True, right=True, which='both')

    clim_ = 0.01, .99

    e_img = norm_img(data[data.shape[0] // 2, :, :, data.shape[3] // 2])
    y_img = norm_img(data[:, data.shape[1] // 2, :, data.shape[3] // 2])
    x_img = norm_img(data[:, :, data.shape[2] // 2, data.shape[3] // 2].T)
    e_plot = img_ax.imshow(e_img, cmap='terrain', aspect='auto', interpolation='gaussian',
                           clim=clim_, )  # origin='lower')
    x_plot = yproj_ax.imshow(x_img, cmap='terrain', aspect='auto', interpolation='gaussian',
                             clim=clim_, )  # origin='lower')
    y_plot = xproj_ax.imshow(y_img, cmap='terrain', aspect='auto', interpolation='gaussian',
                             clim=clim_, )  # origin='lower')

    pe_x = img_ax.axvline(x_range / 2, c='orange')
    pe_y = img_ax.axhline(y_range / 2, c='orange')
    px_x = xproj_ax.axvline(x_range / 2, c='orange')
    px_e = xproj_ax.axhline(e_range / 2, c='orange')
    py_y = yproj_ax.axhline(y_range / 2, c='orange')
    py_e = yproj_ax.axvline(e_range / 2, c='orange')

    def update(e, kx, ky, kz, clim, cmap, binning, interpolate, grid, trackers, trackerscol):
        if normalize:
            e_img = norm_img(data[e, :, :, kz][::binning, ::binning])
            y_img = norm_img(data[:, ky, :, kz][::binning, ::binning])
            x_img = norm_img(data[:, :, kx, kz][::binning, ::binning])
        else:
            e_img = data[e, :, :, kz][::binning, ::binning]
            y_img = data[:, ky, :, kz][::binning, ::binning]
            x_img = data[:, :, kx, kz][::binning, ::binning]
        for axis, plot, img in zip([img_ax, yproj_ax, xproj_ax], [e_plot, x_plot, y_plot], [e_img, x_img.T, y_img]):

            plot.set_data(img)
            plot.set_clim(clim)
            plot.set_cmap(cmap)
            axis.grid(grid)
            if interpolate:
                plot.set_interpolation('gaussian')
            else:
                plot.set_interpolation(None)
            if trackers:
                pe_x.set_xdata(kx)
                pe_x.set_color(trackerscol)
                pe_y.set_ydata(ky)
                pe_y.set_color(trackerscol)
                px_x.set_xdata(kx)
                px_x.set_color(trackerscol)
                px_e.set_ydata(e)
                px_e.set_color(trackerscol)
                py_y.set_ydata(ky)
                py_y.set_color(trackerscol)
                py_e.set_xdata(e)
                py_e.set_color(trackerscol)

    interactive_plot = interactive_output(update, {'e': w_e,
                                                   'kx': w_kx,
                                                   'ky': w_ky,
                                                   'kz': w_kz,
                                                   'clim': w_clim,
                                                   'cmap': w_cmap,
                                                   'binning': w_bin,
                                                   'interpolate': w_interpolate,
                                                   'grid': w_grid,
                                                   'trackers': w_trackers,
                                                   'trackerscol': w_trackercol, });
    display(interactive_plot, tab)
    # display(tab)

    # return fig


def norm_img(data, mode='max'):
    out = np.nan_to_num(data)
    out -= np.amin(out)
    if mode == 'max':
        out /= np.amax(out)
    elif mode == 'mean':
        out /= np.amean(out)
    return out


def det_to_k(x, x0, d):
    ''' Detector pixel coordinate to momentum coordinate (in pixel)'''
    kx = (x - x0)
    while np.abs(kx) > d / 2:
        kx -= np.sign(kx) * d
    return kx


def reduce_to_first_bz(x, x0, reciprocal_lattice_vector):
    kx = (x - x0)
    while np.abs(kx) > reciprocal_lattice_vector / 2:
        kx -= np.sign(kx) * reciprocal_lattice_vector
    return kx


def to_reduced_scheme(x, aoi_k):
    """ obtain momentum coordinate in the reduced scheme.

    Follow Bloch theorem."""
    return aoi_k / 2 - (x + aoi_k / 2) % aoi_k


def to_k_parallel(x, x0, aoi_px, aoi_k):
    return (x - x0) * aoi_k / aoi_px


def to_k_parallel_reduced(x, x0, w_px, w):
    """ Transform pixel to momentum and correct to the reduced brillouin scheme"""
    kx = (x - x0) * w / w_px
    return kx - (kx + w / 2) // w


def to_k_perpendicular(xy, xy0, kf=37.285, aoi_px=219, aoi_k=2.215):
    d = point_distance(xy, xy0)
    px_to_k = aoi_k / aoi_px
    return kf * (1 - np.cos(np.arcsin(d * px_to_k / kf)))


def k_par(px, bz_w_px=219, bz_w=2.215):
    ''' Pixel to momentum converter
    bz_w_px: Brillouin zone width in pixel
    bz_w: Brillouin zone width in inverse Angstroms'''
    return


def d_to_kz(d, kf=37.285, bz_w_px=219, bz_w=2.215):
    ''' Evaluate k_z from Ewald sphere radius
    d: distance in pixels from the center of momentum space on the detector
    kf: photon momentum
    bz_w_px: Brillouin zone width in pixels
    bz_w: Brillouin zone width in inverse Angstroms'''
    return kf * (1 - np.cos(np.arcsin(k_par(d, bz_w_px, bz_w) / kf)))


def point_distance(pt1, pt2):
    return np.sqrt(np.power(pt2[0] - pt1[0], 2) + np.power(pt2[1] - pt1[1], 2))


def get_k_xyz(pt, k_center=(179, 737), reciprocal_lattice_vector=(218, 218), reciprocal_unit_cell=(2.215, 2.215, 1),
              kf=37.285):
    """ return the kx,ky,kz coordinates of a point.
    param:
        pt: tuple(float,float)
            point to evaluate (row,col)
        k_center:
            center of momentum space
        reciprocal_lattice_vector: float
            distance between 2 gamma points - BZ width
        kf: float
            final momentum from energy of photoemission photon

    given distance from gamma point and size of the brillouin zone"""
    y_c, x_c = k_center
    y_w, x_w = reciprocal_lattice_vector
    y, x = pt

    kx = k_par(det_to_k(x, x_c, reciprocal_lattice_vector[1]), bz_w_px=reciprocal_lattice_vector[1],
               bz_w=reciprocal_unit_cell[1])
    ky = k_par(det_to_k(y, y_c, reciprocal_lattice_vector[0]), bz_w_px=reciprocal_lattice_vector[0],
               bz_w=reciprocal_unit_cell[0])
    d = point_distance(pt, k_center)
    kz = d_to_kz(d, kf=kf, bz_w_px=np.mean(reciprocal_lattice_vector))
    return kx, ky, kz


def slice_to_ev(ToF, ToF_to_ev, t0):
    return (ToF - t0) * ToF_to_ev



# -*- coding: utf-8 -*-

"""
@author: Steinn Ymir Agustsson
"""

import os
import sys
import h5py


try:
    from camp.pah.beamtimedaqaccess import BeamtimeDaqAccess as _BeamtimeDaqAccess, H5FileDataAccess as _H5FileDataAccess, \
    H5FileManager as _H5FileManager

except ModuleNotFoundError:
    import configparser

    settings = configparser.ConfigParser()  # TODO: find a smarter way
    if os.path.isfile(os.path.join(os.path.dirname(__file__), 'SETTINGS.ini')):
        settings.read(os.path.join(os.path.dirname(__file__), 'SETTINGS.ini'))
    else:
        settings.read(os.path.join(os.path.dirname(os.path.dirname(__file__)), 'SETTINGS.ini'))
    sys.path.append(settings['paths']['PAH_MODULE_DIR'])

    from camp.pah.beamtimedaqaccess import BeamtimeDaqAccess as _BeamtimeDaqAccess, H5FileDataAccess as _H5FileDataAccess, \
    H5FileManager as _H5FileManager


# Below are the redefined classes belonging to PAH that should correct the
# problems induced by adding the macrobunchID information to the data.


class BeamtimeDaqAccess(_BeamtimeDaqAccess):
    """ Overwriting original class to apply local corrections to PAH code.
    """

    def __init__(self, fileAccess):
        super(BeamtimeDaqAccess, self).__init__(fileAccess)

    @staticmethod
    def create(rootDirectoryOfH5Files):
        """ Creates a BeamtimeDaqAccess object for the given root directory - API

        **Parameter**\n
        rootDirectoryOfH5Files (str): The root directory of the HDF files. The root
                directory contains sub directories for different FLASH DAQ streams.

        **Return**\n
        BeamtimeDaqAccess: The ready to use beamtime DAQ access object.

        **Raise**\n
        AssertionError: If the given rootDirectoryOfH5Files does not exist.
        """

        fileAccess = H5FileDataAccess(H5FileManager(rootDirectoryOfH5Files))
        return BeamtimeDaqAccess(fileAccess)


class H5FileDataAccess(_H5FileDataAccess):
    """ Wrapper class for correcting PAH code for defining valid channels to read from.
    """

    def __init__(self, h5FileManager):
        super(H5FileDataAccess, self).__init__(h5FileManager)
        # super(self).__init__(h5FileManager)

    def isValidChannel(self, channelName):
        """ Define the valid channels in HDF5 data.

        Add the timing section of the hdf5 dataset as valid channel, as the
        timing channel was not considered a valid channel in the original
        PAH code.

        **Return**\n
        (bool): True: channel is valid, False: channel not valid.
        """
        return channelName in self.allChannelNames() \
               or channelName.startswith('/uncategorised/') \
               or channelName.startswith('/FL2/') \
               or channelName.startswith('/FL1/') \
               or channelName.startswith('/Experiment/') \
               or channelName.startswith('/Photon Diagnostic/') \
               or channelName.startswith('/Electron Diagnostic/') \
               or channelName.startswith('/Beamlines/') \
               or channelName.startswith('/Timing/')  # <--for datasets before 08-2018

    def filledInDesiredDataSets(self, sortedDesiredDataSets):
        assert sortedDesiredDataSets, "Precond.: desiredDataSets not empty"

        # finding largest datumsShape
        largestShape = 0
        for currentDataset in sortedDesiredDataSets:
            with h5py.File(currentDataset.fileMeta.fileName(), 'r') as h5file:
                if h5file[currentDataset.channelName].shape[1] > largestShape:
                    largestDataset = currentDataset
                    largestShape = h5file[currentDataset.channelName].shape[1]
        largestDataset.desiredDatasetInitializedFromFile()

        # all the result parts need to be the same shape. Therefore, they need to be the largest possible
        result= [sortedDesiredDataSets[0].desiredDatasetInitializedFromSample(largestDataset)]
        preceedingDataset= result[0]

        for currentDataset in sortedDesiredDataSets[1:]:
            if currentDataset.pulseIdInterval[0] - preceedingDataset.pulseIdInterval[1] > 0:
                nanDataSet= _H5FileDataAccess.NaNDesiredDataSet((preceedingDataset.pulseIdInterval[1], currentDataset.pulseIdInterval[0]), preceedingDataset)
                result.append(nanDataSet)
            result.append(currentDataset.desiredDatasetInitializedFromSample(preceedingDataset))
            preceedingDataset= currentDataset
        assert self.hasNoGapsBetween(result), "Postcond.: No gaps between file chunks."
        return result


class H5FileManager(_H5FileManager):
    """ Wrapper for pointing to original class in  PAH module.
    """

    def __init__(self, rootDirectoryOfH5Files):
        super(H5FileManager, self).__init__(rootDirectoryOfH5Files)



# -*- coding: utf-8 -*-
"""

@author: Steinn Ymir Agustsson
"""

from .misc import repr_byte_size
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

import dask
from dask.diagnostics import ProgressBar


def channel_report(dd):
    """ Generates a table containing relevant statistical quantities on all available channels [EXPENSIVE!!]

    Args:
        dd (dask.DataFrame): Table to analyse

    Returns:
        df (pandas.DataFrame): table with computed values
    Notes:
        Author: Steinn Ymir Agustsson <sagustss@uni-mainz.de>

    """
    print('creating channel report...')
    with ProgressBar():
        chan_dict = {}
        for chan in dd.columns:
            vals = dd[chan].compute()
            clean = vals[np.isfinite(vals)]
            chan_dict[chan] = {}
            chan_dict[chan]['min'] = min(clean)
            chan_dict[chan]['max'] = max(clean)
            chan_dict[chan]['amp'] = chan_dict[chan]['max'] - chan_dict[chan]['min']
            chan_dict[chan]['mean'] = np.mean(clean)
            chan_dict[chan]['std'] = np.std(clean)
            chan_dict[chan]['len'] = len(vals)
            chan_dict[chan]['len_nan'] = len(vals) - len(clean)
            chan_dict[chan]['max_loc'] = vals.idxmax()
            chan_dict[chan]['min_loc'] = vals.idxmin()
            


    df = pd.DataFrame.from_dict(chan_dict).T
    return df

def channel_statistics(dd,columns=None):
    """ Statistical overiview of the dataset.

    The report contains minimum, maximum, amplitude(max-min), mean and standard
    deviation of each column.

    Args:
        dd (dask.DataFrame): Table to analyse
        columns (list,None): list of columns to analyse, if None it checks all
            columns in the dataframe.
    Returns:
        df (pandas.DataFrame): table with computed values
    Notes:
        Author: Steinn Ymir Agustsson <sagustss@uni-mainz.de>

    """
    print('creating channel report...')
    chan_dict = {}
    if columns is None:
        columns = dd.columns
    for chan in columns:
        try:
            vals = dd[chan]
            chan_dict[chan] = {}
            chan_dict[chan]['min'] = vals.min()
            chan_dict[chan]['max'] = vals.max()
            chan_dict[chan]['amp'] = chan_dict[chan]['max'] - chan_dict[chan]['min']
            chan_dict[chan]['mean'] = vals.mean()
            chan_dict[chan]['std'] = vals.std()
#             chan_dict[chan]['len'] = vals.size
#             chan_dict[chan]['len_nan'] = vals.size - vals.isna().size

        except KeyError:
            print(f'No column {chan} in dataframe')

    with ProgressBar():
        d, = dask.compute(chan_dict)
    return pd.DataFrame.from_dict(d)

def binned_array_size(processor):
    """ Prints the expected size in memory of the binned array, computed with
    the current binning parameters"""

    voxels = np.prod([len(x) for x in processor.binAxesList])
    print(repr_byte_size(voxels.astype(np.float64) * 64.))


def plot_channels(processor):
    """ Generates a plot of each available channel.

    Useful for diagnostics on data conditions

    Args:
        processor (DldProcessor): processor instance

    """
    cols = processor.dd.columns
    f, ax = plt.subplots(len(cols) // 2, 2, figsize=(15, 20))
    for i, chan in enumerate(processor.dd.columns):
        vals = processor.dd[chan].compute()
        clean = vals[np.isfinite(vals)]
        start, stop, mean = clean.min(), clean.max(), clean.mean()

        if np.abs((start - mean) / (stop - mean)) > 5:
            start = stop - 2 * mean
        elif np.abs((stop - mean) / (start - mean)) > 5:
            stop = start + 2 * mean

        step = max((stop - start) / 1000, 1)
        x = processor.addBinning(chan, start, stop, step)
        res = processor.computeBinnedData()
        ax[i // 2, i % 2].set_title(chan, fontsize='x-large')
        if len(res) > 1:
            try:
                ax[i // 2, i % 2].plot(x, res)
            except ValueError:
                ax[i // 2, i % 2].plot(x, res[:-1])
        processor.resetBins()


def plot_GMD_vs_bunches(self):
    """ not sure what this was done for...
    """

    f, ax = plt.subplots(1, 2)
    self.resetBins()
    ubId = self.addBinning('dldMicrobunchId', 1, 500, 1)
    result_ubId = self.computeBinnedData()
    ax[0].plot(ubId, result_ubId / max(result_ubId), label='Photoelectron count')
    ax[0].set_title('Microbunch')

    self.resetBins()

    MbId_values = self.dd['macroBunchPulseId'].compute()
    clean = MbId_values[np.isfinite(MbId_values)]
    start, stop, mean = clean.min(), clean.max(), clean.mean()
    if np.abs((start - mean) / (stop - mean)) > 5:
        start = stop - 2 * mean
    elif np.abs((stop - mean) / (start - mean)) > 5:
        stop = start + 2 * mean
    step = max((stop - start) / 1000, 1)

    MbId = self.addBinning('macroBunchPulseId', start, stop, step)
    result_MbId = self.computeBinnedData()
    ax[1].plot(MbId, result_MbId / max(result_MbId), label='Photoelectron count')
    ax[1].set_title('Macrobunch')

    gmd_values = np.nan_to_num(self.dd['gmdBda'].compute())
    ubId_values = self.dd['dldMicrobunchId'].compute()

    gmd_ubId = np.zeros_like(ubId)
    for g, ub in zip(gmd_values, ubId_values):
        ub_idx = int(ub)
        if 0 < ub_idx < len(gmd_ubId):
            gmd_ubId[ub_idx] += g

    gmd_MbId = np.zeros_like(MbId)
    for g, Mb in zip(gmd_values, MbId_values):
        Mb_idx = int(Mb)
        if 0 < Mb_idx < len(gmd_MbId):
            gmd_MbId[Mb_idx] += g

    ax[0].plot(ubId, gmd_ubId / max(gmd_ubId), label='GMD')
    ax[1].plot(MbId, gmd_MbId / max(gmd_MbId) + 1, label='GMD')
    for a in ax:
        a.legend()
        a.grid()



# -*- coding: utf-8 -*-
"""

@author: Steinn Ymir Agustsson, Davide Curcio, Maciej Dendzik
"""
import sys, os
import numpy as np


def gen_sector_correction(prc, energies, eref, tofVoltage=None, sampleBias=None, monoEnergy=None):
    """
    This function is helpful in generating the sector_correction list.
    This takes into account the time shift caused by the bit stealing hack
    plus is keeps track of the time shift due to detector misalignment by making sure
    all values of energies are at eref.

    Usage: use the function to create the sector_correction list and assign it to prc.SECTOR_CORRECTION
    or paste it into the settings with no brackets

    :param prc:
    :param energies:
    :param eref:
    :return:
    """

    if sampleBias is None:
        sampleBias = np.nanmean(prc.dd['sampleBias'].values)
    if monoEnergy is None:
        monoEnergy = np.nanmean(prc.dd['monochromatorPhotonEnergy'].values)
    if tofVoltage is None:
        tofVoltage = np.nanmean(prc.dd['tofVoltage'].values)

    # Here, the most basic sector_correction list is generated, where the bit stealing hack gets corrected
    n_sectors=prc.dd['dldSectorId'].values.compute().max().astype(int)+1
    sector_correction = np.floor(np.array(range(n_sectors))*np.power(2,prc.DLD_ID_BITS)/n_sectors)
    # e.g. for 3 stolen bits, 8 detector ids (new s8 data), this will look like [0,1,2,3,4,5,6,7]
    # e.g. for 1 stolen bit, 8 detector ids (old modified s8 data), this will look like [0,0,0,0,1,1,1,1]
    # e.g. for 1 stolen bit, 2 detector ids (old s8 data), this will look like [0,1]


    t_ref = energy2tof(eref, l=0.965, eoffset=-2.64-sampleBias+tofVoltage+monoEnergy)

    times=[]
    for ee in energies:
        times.append((energy2tof(ee, l=0.965, eoffset=-2.64-47+30+monoEnergy)-t_ref)/prc.TOF_STEP_TO_NS)

    sector_correction=sector_correction+np.array(times)
    # here the sectors are shifter by the required time to align the energies

    return sector_correction



# %% Energy calibration

""" The following functions convert between binding energy (Eb) in eV (negative convention)
    and time of flight (ToF) in ns.

    The formula used is based on the ToF for an electron with a kinetic energy Ek. Then the
    binding energy Eb is given by
    
    .. math::

        -E_b = E_k + W - hv - V = \\frac{1}{2} mv^2 + W - hv - V

    With W the work function, hv the photon energy, V the electrostatic potential applied to
    the sample, v the velocity of the electrons in the drift tube, m the mass of the electron.
    The velocity v in the drift tube can be calculated knowing the length (1m) and the flight
    time in the drift tube. The measured ToF, however, has some offset due to clock start not
    coinciding with entry in the drift section.

    offs is supposed to include the time offset for when the electrons enter the drift section.
    Its main mainly affects peak spacing, and there are several strategies for calibrating this
    value:
    1.  By measuring the photon peak and correcting by some extractor voltage-dependent offset
    2.  Shifting the potential by 1V and imposing the same shift in the measured spectrum
    3.  Imposing some calibrated spacing between features in a spectrum

    oo is supposed to include -W+hv+V. It mainly affects absolute position of the peaks, and
    there are several strategies for calibrating this value:
    1.  By getting the correct values for W, hv, and V
    2.  It can be calibrated by imposing peak position

Parameters:
    t (float) the ToF
    e (float) the binding energy

"""


def tof2energy(t, toffset=None, eoffset=None, l=None):
    """ Transform ToF to eV.

    The functions (tof2energy and energy2tof) convert between binding energy (:math:`E_b`) in eV (negative convention)
    and time of flight (ToF) in ns.

    The formula used is based on the ToF for an electron with a kinetic energy :math:`E_k`. Then the
    binding energy :math:`E_b` is given by

    .. math::

        -E_b = E_k + W - hv - V = \\frac{1}{2} mv^2 + W - hv - V


    With W the work function, hv the photon energy, V the electrostatic potential applied to
    the sample with respect to the drift section voltage, v the velocity of the electrons in the drift tube,
    m the mass of the electron.

    The velocity v in the drift tube can be calculated knowing the length (1m) and the flight
    time in the drift tube. The measured ToF, however, has some offset due to clock start not
    coinciding with entry in the drift section.

    toffset is supposed to include the time offset for when the electrons enter the drift section.
    Its main mainly affects peak spacing, and there are several strategies for calibrating this value,

    1.  By measuring the photon peak and correcting by some extractor voltage-dependent offset
    2.  Shifting the potential by 1V and imposing the same shift in the measured spectrum
    3.  Imposing some calibrated spacing between features in a spectrum

    eoffset is supposed to include -W+hv+V. It mainly affects absolute position of the peaks, and
    there are several strategies for calibrating this value,

    1.  By getting the correct values for W, hv, and V
    2.  It can be calibrated by imposing peak position

    Parameters:
        t: float
            The time of flight
        toffset: float
            The time offset from thedld clock start to when the fastest photoelectrons reach the detector
        eoffset: float
            The energy offset given by W-hv-V
        l: float
            the effective length of the drift section

    Return:
        e: float
            The binding energy

    Authors:
        Davide Curcio <davide.curcio@phys.au.dk>
    """

    from configparser import ConfigParser
    settings = ConfigParser()
    if os.path.isfile(os.path.join(os.path.dirname(__file__), 'SETTINGS.ini')):
        settings.read(os.path.join(os.path.dirname(__file__), 'SETTINGS.ini'))
    else:
        settings.read(
            os.path.join(
                os.path.dirname(os.path.dirname(os.path.dirname(__file__))), 'SETTINGS.ini'))

    if toffset is None:
        toffset = float(settings['processor']['ET_CONV_T_OFFSET'])
    if eoffset is None:
        eoffset = float(settings['processor']['ET_CONV_E_OFFSET'])
    if l is None:
        l = float(settings['processor']['ET_CONV_L'])
    e = 0.5 * 1e18 * 9.10938e-31 * l * l / (((t) - toffset) * ((t) - toffset)) / 1.602177e-19 - eoffset
    return e


def energy2tof(e, toffset=None, eoffset=None, l=None):
    """ Transform eV to time of flight (ToF).

    The functions (tof2energy and energy2tof) convert between binding energy (:math:`E_b`) in eV (negative convention)
    and time of flight (ToF) in ns.

    The formula used is based on the ToF for an electron with a kinetic energy :math:`E_k`. Then the
    binding energy :math:`E_b` is given by

    .. math::

        -E_b = E_k + W - hv - V = \\frac{1}{2} mv^2 + W - hv - V


    With W the work function, hv the photon energy, V the electrostatic potential applied to
    the sample, v the velocity of the electrons in the drift tube, m the mass of the electron.
    The velocity v in the drift tube can be calculated knowing the length (1m) and the flight
    time in the drift tube. The measured ToF, however, has some offset due to clock start not
    coinciding with entry in the drift section.

    offs is supposed to include the time offset for when the electrons enter the drift section.
    Its main mainly affects peak spacing, and there are several strategies for calibrating this value,

    1.  By measuring the photon peak and correcting by some extractor voltage-dependent offset
    2.  Shifting the potential by 1V and imposing the same shift in the measured spectrum
    3.  Imposing some calibrated spacing between features in a spectrum

    eoffset is supposed to include -W+hv+V. It mainly affects absolute position of the peaks, and
    there are several strategies for calibrating this value,

    1.  By getting the correct values for W, hv, and V
    2.  It can be calibrated by imposing peak position

    Parameters:
        e: float
            The binding energy
        toffset: float
            The time offset from thedld clock start to when the fastest photoelectrons reach the detector
        eoffset: float
            The energy offset given by W-hv-V
        l: float
            the effective length of the drift section

    Return:
        t: float
            The time of flight

    Authors:
        Davide Curcio <davide.curcio@phys.au.dk>
    """

    from configparser import ConfigParser
    settings = ConfigParser()
    if os.path.isfile(os.path.join(os.path.dirname(__file__), 'SETTINGS.ini')):
        settings.read(os.path.join(os.path.dirname(__file__), 'SETTINGS.ini'))
    else:
        settings.read(
            os.path.join(
                os.path.dirname(os.path.dirname(os.path.dirname(__file__))), 'SETTINGS.ini'))

    if toffset is None:
        toffset = float(settings['processor']['ET_CONV_T_OFFSET'])
    if eoffset is None:
        eoffset = float(settings['processor']['ET_CONV_E_OFFSET'])
    if l is None:
        l = float(settings['processor']['ET_CONV_L'])
    t = l * np.sqrt(0.5 * 1e18 * 9.10938e-31 / 1.602177e-19 / (e + eoffset)) + toffset
    return t


def e2t(e, toffset=None, eoffset=None, l=0.77):
    """ Same as energy2tof, but different name for retrocompatibility

    Parameters:
        e: float
            The binding energy
        toffset: float
            The time offset from thedld clock start to when the fastest photoelectrons reach the detector
        eoffset: float
            The energy offset given by W-hv-V
        l: float
            the effective length of the drift section

    Return:
        t: float
            The time of flight

    Authors:
        Davide Curcio <davide.curcio@phys.au.dk>
    """

    return energy2tof(e, toffset, eoffset, l)


def t2e(t, toffset=None, eoffset=None, l=0.77):
    """ Same as tof2energy, but different name for retrocompatibility

    Parameters:
        t: float
            The time of flight
        toffset: float
            The time offset from thedld clock start to when the fastest photoelectrons reach the detector
        eoffset: float
            The energy offset given by W-hv-V
        l: float
            the effective length of the drift section

    Return:
        e: float
            The binding energy

    Authors:
        Davide Curcio <davide.curcio@phys.au.dk>
    """

    return tof2energy(t, toffset, eoffset, l)


# %% Detector calibration

# ==================
# Methods by Mac! Built for the April 2018 beamtime, where a four quadrant detector was used.
# ==================

def shiftQuadrants(self, shiftQ1=0.231725, shiftQ2=-0.221625, shiftQ3=0.096575, shiftQ4=-0.106675, xCenter=1350,
                   yCenter=1440):
    """ Apply corrections to the dataframe. (Maciej Dendzik)

    Each quadrant of DLD is shifted in DLD time by shiftQn.
    xCenter and yCenter are used to define the center of the division.
    
    +-----------+-----------+
    |    Q2     |     Q4    |
    +-----------+-----------+
    |    Q1     |     Q3    |
    +-----------+-----------+

    This picture is upside-down in ``plt.imshow`` because it starts from 0 in top right corner.
    """
    # Q1
    # daskdataframe.where(condition,value) keeps the data where condition is True
    # and changes them to value otherwise.
    cond = ((self.dd['dldPosX'] > xCenter) | (self.dd['dldPosY'] > yCenter))
    self.dd['dldTime'] = self.dd['dldTime'].where(cond, self.dd['dldTime'] + shiftQ1)
    cond = ((self.dd['dldPosX'] > xCenter) | (self.dd['dldPosY'] < yCenter))
    self.dd['dldTime'] = self.dd['dldTime'].where(cond, self.dd['dldTime'] + shiftQ2)
    cond = ((self.dd['dldPosX'] < xCenter) | (self.dd['dldPosY'] > yCenter))
    self.dd['dldTime'] = self.dd['dldTime'].where(cond, self.dd['dldTime'] + shiftQ3)
    cond = ((self.dd['dldPosX'] < xCenter) | (self.dd['dldPosY'] < yCenter))
    self.dd['dldTime'] = self.dd['dldTime'].where(cond, self.dd['dldTime'] + shiftQ4)


def filterCircleDLDPos(self, xCenter=1334, yCenter=1426, radius=1250):
    """ Apply corrections to the dataframe. (Maciej Dendzik)

    Filters events with dldPosX and dldPosY within the radius from (xCenter,yCenter).

    """

    self.dd = self.dd[
        (((self.dd['dldPosX'] - xCenter) ** 2 + (self.dd['dldPosY'] - yCenter) ** 2) ** 0.5 <= radius)]


def correctOpticalPath(self, poly1=-0.00020578, poly2=4.6813e-7, xCenter=1334, yCenter=1426):
    """ Apply corrections to the dataframe. (Maciej Dendzik)

    Each DLD time is subtracted with a polynomial poly1*r + poly2*r^2,
    where r=sqrt((posx-xCenter)^2+(posy-yCenter)^2)

    This function makes corrections to the time of flight which take into account
    the path difference between the center of the detector and the edges of the detector.

    """
    # Q1
    # daskdataframe.where(condition,value) keeps the data where condition is True
    # and changes them to value otherwise.

    self.dd['dldTime'] = self.dd['dldTime'] - \
                         (poly1 * ((self.dd['dldPosX'] - xCenter) ** 2 + (
                                 self.dd['dldPosY'] - yCenter) ** 2) ** 0.5 + \
                          poly2 * ((self.dd['dldPosX'] - xCenter) ** 2 + (self.dd['dldPosY'] - yCenter) ** 2))
    


# -*- coding: utf-8 -*-

'''
@author: Markus Scholz, Mario Reiser\n
Masking tool for image data
'''

from matplotlib import pyplot as plt
import matplotlib as mpl
import numpy as np

import matplotlib.path as path
from matplotlib.colors import ListedColormap

from time import sleep
import ipywidgets as widgets

def masker(data, mask1=[]):
    

    if not isinstance(mask1, np.ndarray):
        mask1 = np.ones_like(data)
    
    xs = []
    ys = []
    mask = []
    
    def invert_mask(*args):
        ind = mask1 > 0
        mask1[ind] = 0
        mask1[~ind] = 1
        mask.set_data(mask1)

    def reset_mask(*args):
        mask1[:] = 1
        mask.set_data(mask1)

    def save_mask(*args):
        savname = savwid.value
        savname = savname.replace('.npy', "") + '.npy'
        np.save(savname, mask1.astype(bool))

    def cancel_masking(*args):
        del xs[:]
        del ys[:]
        line.set_data(xs, ys)
        line.figure.canvas.draw_idle()
        
    def mask_roi(*args):
        lx,ly=np.shape(mask1)
        x, y = np.meshgrid(np.arange(lx), np.arange(ly))
        x, y = x.flatten(), y.flatten()
        points = np.vstack((x,y)).T
        xy = np.vstack((np.array(ys),np.array(xs))).T
        new_submask = ~path.Path(xy).contains_points(points).reshape(ly,lx).T 
        np.multiply(mask1, new_submask, out=mask1)
        del xs[:]
        del ys[:]
        mask.set_data(mask1.astype(int))

    def mask_outside(*args):
        lx,ly=np.shape(mask1)
        x, y = np.meshgrid(np.arange(lx), np.arange(ly))
        x, y = x.flatten(), y.flatten()
        points = np.vstack((x,y)).T
        xy = np.vstack((np.array(ys),np.array(xs))).T
        new_submask = path.Path(xy).contains_points(points).reshape(ly,lx).T 
        np.multiply(mask1, new_submask, out=mask1)
        del xs[:]
        del ys[:]
        mask.set_data(mask1.astype(int))

    def unmask_roi(*args):
        lx,ly=np.shape(mask1)
        x, y = np.meshgrid(np.arange(lx), np.arange(ly))
        x, y = x.flatten(), y.flatten()
        points = np.vstack((x,y)).T
        xy = np.vstack((np.array(ys),np.array(xs))).T
        new_submask = path.Path(xy).contains_points(points).reshape(ly,lx).T 
        mask1[new_submask] = True
        del xs[:]
        del ys[:]
        mask.set_data(mask1.astype(int))

    def onclick(event,line):
        xs.append(event.xdata)
        ys.append(event.ydata)

        line.set_data(xs, ys)
        line.figure.canvas.draw_idle()
            
    def onkey(event,line):
        if event.key == 'c':
            cancel_masking()
        elif event.key == 'm':
            mask_roi()
        elif event.key == 'o':
            mask_outside()
        elif event.key == 'u':
            unmask_roi()
        elif event.key == 'i':
            invert_mask()

    savwid = widgets.Text(
        value='example_mask.npy',
        placeholder='type here',
        description='Filename:',
        disabled=False
    )
    mvwid = widgets.Text(
        value='0',
        description='mask value:',
        placeholder='type herer',
        disabled=False
    )
    row1 = widgets.HBox([savwid, mvwid])

    invert_button = widgets.Button(
        description='invert (i)',
        disabled=False,
        tooltip='invert the complete mask',
        icon='',
    )
    invert_button.on_click(invert_mask)
    reset_button = widgets.Button(
        description='reset',
        disabled=False,
        tooltip='all vales set to default',
    )
    reset_button.on_click(reset_mask)
    save_button = widgets.Button(
        description='save mask',
        disabled=False,
        tooltip='save mask to npy file',
    )
    save_button.on_click(save_mask)
    row3 = widgets.HBox([invert_button,
                         reset_button,
                         save_button])

    mask_button = widgets.Button(
        description='mask (m)',
        disabled=False,
        tooltip='mask roi',
        icon='',
    )
    mask_button.on_click(mask_roi)
    unmask_button = widgets.Button(
        description='unmask (u)',
        disabled=False,
        tooltip='unmask roi',
    )
    unmask_button.on_click(unmask_roi)
    outside_button = widgets.Button(
        description='mask outside (o)',
        disabled=False,
        tooltip='mask everything outside the roi',
    )
    outside_button.on_click(mask_outside)
    cancel_button = widgets.Button(
        description='cancel (c)',
        disabled=False,
        tooltip='cancel roi',
    )
    cancel_button.on_click(cancel_masking)
    row2 = widgets.HBox([mask_button,
                         unmask_button,
                         outside_button,
                         cancel_button])

    out = widgets.Output()
    display(out)
    with out:

        display(widgets.VBox([row1,row2,row3]))

        fig = plt.figure(figsize=(8,8))
        ax = plt.axes(position=[.06, .05, .94, .94])
        plt.ion()

        cmap = mpl.colors.ListedColormap(['orange', 'white'])

        # Get the colormap colors
        my_cmap = cmap(np.arange(cmap.N))

        # Set alpha
        my_cmap[:,-1] = 0.7

        # Create new colormap
        cmap = ListedColormap(my_cmap)
        cmap.set_over('orange', alpha=0)

        bounds=[0,.5,1]
        norm = mpl.colors.BoundaryNorm(bounds, cmap.N)
        ax.imshow(data, interpolation='none', norm=mpl.colors.LogNorm(), aspect='auto')
        mask = plt.imshow(mask1.astype(int), interpolation='none', norm=norm, cmap=cmap, aspect='auto')
        line, = ax.plot([],[],'.-k')
        cid_m = fig.canvas.mpl_connect('button_press_event', lambda event: onclick(event,line))
        cid_k = fig.canvas.mpl_connect('key_press_event', lambda event: onkey(event,line))
        fig.canvas.draw()
    
    
